package BooleanFormula.BooleanFormula;

import JSON.structures.Element;
import JSON.structures.ElementReference;
import lists.DynamicArrayNumbers.Structures.DynamicArrayNumbers;
import references.references.BooleanArrayReference;
import references.references.BooleanReference;
import references.references.NumberReference;
import references.references.StringArrayReference;

import static JSON.parser.parser.ReadJSON;
import static arrays.arrays.arrays.StringsEqual;
import static java.lang.Math.*;
import static lists.DynamicArrayNumbers.DynamicArrayNumbersFunctions.DynamicArrayNumbersFunctions.*;
import static references.references.references.CreateBooleanReference;

public class BooleanFormula {
	public static boolean Evaluate(ConjunctiveNormalForm cnf, boolean [] assignments){
		boolean cnfEval, clauseEval;
		double i, j, v;
		boolean s;
		Clause c;

		cnfEval = true;
		for(i = 0d; i < cnf.clauses.length; i = i + 1d){
			c = cnf.clauses[(int)i];

			clauseEval = true;

			for(j = 0d; j < c.literals.length; j = j + 1d){
				v = abs(c.literals[(int)j]) - 1d;
				s = c.literals[(int) j] > 0d;

				if(j == 0d){
					if (s) {
						clauseEval = assignments[(int)v];
					} else {
						clauseEval = !assignments[(int)v];
					}
				}else {
					if (s) {
						clauseEval = clauseEval || assignments[(int)v];
					} else {
						clauseEval = clauseEval || !assignments[(int)v];
					}
				}
			}

			cnfEval = cnfEval && clauseEval;
		}


		return cnfEval;
	}

	public static boolean JSONToCNF(ConjunctiveNormalForm cnf, char [] json){
		ElementReference elementRef;
		StringArrayReference errors;
		boolean success;
		double i, j;
		Clause c;
		Element e, p, l;

		elementRef = new ElementReference();
		errors = new StringArrayReference();

		success = ReadJSON(json, elementRef, errors);

		if(success){
			e = elementRef.element;

			if(StringsEqual(e.type, "array".toCharArray())){
				cnf.clauses = new Clause[e.array.length];

				for(i = 0d; i < e.array.length && success; i = i + 1d){
					c = new Clause();
					p = e.array[(int)i];

					if(StringsEqual(p.type, "array".toCharArray())){
						c.literals = new double[p.array.length];

						for(j = 0d; j < p.array.length && success; j = j + 1d) {
							l = p.array[(int)j];
							if(StringsEqual(l.type, "number".toCharArray())){
								c.literals[(int)j] = l.number;
							}else{
								success = false;
							}
						}

					}else{
						success = false;
					}

					cnf.clauses[(int)i] = c;
				}

			}else{
				success = false;
			}
		}

		return success;
	}

	public static ArrayOfArraysOfNumbers CreateArrayOfArraysOfNumbers(){
		ArrayOfArraysOfNumbers arr;

		arr = new ArrayOfArraysOfNumbers();

		arr.data = new DynamicArrayNumbers[0];
		arr.size = 0d;
		arr.capacity = 0d;

		return arr;
	}

	public static ArrayOfImplicationNodes CreateArrayOfImplicationNodes(){
		ArrayOfImplicationNodes arr;

		arr = new ArrayOfImplicationNodes();

		arr.data = new ImplicationNode[0];
		arr.size = 0d;
		arr.capacity = 0d;

		return arr;
	}

	public static void ArrayOfArraysOfNumbersAdd(ArrayOfArraysOfNumbers arr, DynamicArrayNumbers nums) {
		DynamicArrayNumbers [] temp;
		double i, n;

		if (arr.capacity == arr.size) {
			temp = arr.data;

			n = arr.capacity * 2d;
			if (n < 8d) {
				n = 8d;
			}

			arr.data = new DynamicArrayNumbers[(int) n];

			for (i = 0d; i < arr.size; i = i + 1d) {
				arr.data[(int) i] = temp[(int) i];
			}

			delete(temp);

			arr.capacity = n;
		}

		arr.data[(int) arr.size] = nums;
		arr.size = arr.size + 1d;
	}

	public static void ArrayOfImplicationNodesAdd(ArrayOfImplicationNodes arr, ImplicationNode node) {
		ImplicationNode[] temp;
		double i, n;

		if (arr.capacity == arr.size) {
			temp = arr.data;

			n = arr.capacity * 2d;
			if (n < 8d) {
				n = 8d;
			}

			arr.data = new ImplicationNode[(int) n];

			for (i = 0d; i < arr.size; i = i + 1d) {
				arr.data[(int) i] = temp[(int) i];
			}

			delete(temp);

			arr.capacity = n;
		}

		arr.data[(int) arr.size] = node;
		arr.size = arr.size + 1d;
	}

	public static void ArrayOfImplicationNodesRemoveAt(ArrayOfImplicationNodes arr, double index) {
		double i;

		if (index >= 0d && index < arr.size) {
			arr.size = arr.size - 1d;
			for (i = index; i < arr.size; i = i + 1d) {
				arr.data[(int) i] = arr.data[(int) (i + 1d)];
			}
			// We don't delete here because we use the implication nodes
			// as pointers and we don't "new" new nodes every time.
			// Very specific but whatever...
		} else {
			// System.out.println("Attempting to do bad stuff");
		}
	}

	public static double GetNumberOfVariables(ConjunctiveNormalForm cnf) {
		double i, j, upperBound, totalVars;
		Clause cl;
		boolean[] unique;

		upperBound = 0d;

		for (i = 0d; i < cnf.clauses.length; i = i + 1d) {
			cl = cnf.clauses[(int) i];
			upperBound = upperBound + cl.literals.length;
		}

		unique = new boolean[(int) upperBound];
		for (i = 0d; i < cnf.clauses.length; i = i + 1d) {
			cl = cnf.clauses[(int) i];
			for (j = 0d; j < cl.literals.length; j = j + 1d) {
				unique[(int) (abs(cl.literals[(int) j]) - 1d)] = true;
			}
		}

		totalVars = 0d;

		for (i = 0d; i < upperBound; i = i + 1d) {
			if (unique[(int) i]) {
				totalVars = totalVars + 1d;
			}
		}

		delete(unique);

		return totalVars;
	}

	public static boolean[] AssignmentsToBooleanArray(double[] assignments) {
		boolean[] result;
		double i;

		result = new boolean[assignments.length];

		for (i = 0d; i < assignments.length; i = i + 1d) {
			result[(int) i] = assignments[(int) i] == 1d;
		}
		return result;
	}

	public static boolean AreAllVariablesAssigned(SolveState s) {
		boolean success, toBreak;
		double i;

		success = true;
		toBreak = false;

		for (i = 0d; i < s.assignments.length && !toBreak; i = i + 1d) {
			if (s.assignments[(int) i] == -1d) {
				toBreak = true;
				success = false;
			}
		}
		return success;
	}

	public static double ComputeClause(SolveState s, double[] literals, double literalsSize) {
		double i, lit, v, result;
		boolean toBreak;

		toBreak = false;

		result = -100d;

		for (i = 0d; i < literalsSize && !toBreak; i = i + 1d) {
			lit = literals[(int) i];

			v = ComputeValue(s, lit);

			if (v == -1d) {
				result = -1d;
				toBreak = true;
			} else {
				result = max(result, v);
			}
		}

		return result;
	}

	public static double ComputeValue(SolveState s, double literal) {
		double result;

		result = s.assignments[(int) (abs(literal) - 1d)];
		if (result != -1d) {
			if (literal < 0d) {
				// Flip true/false if the literal in the clause is negated
				if (result == 1d) {
					result = 0d;
				} else {
					result = 1d;
				}
			}
		}

		return result;
	}

	// Checks if clause is a unit clause. If and only if there is
	// exactly 1 literal unassigned, and all the other literals having
	// value of 0.
	// :param clause: set of ints
	// :returns: (is_clause_a_unit, the_literal_to_assign, the clause)
	//
	// We have a parameter "clauseSize" because the learnt clauses are
	// arrays of numbers, which means that clause.length actually returns
	// the capacity and not the number of valid elements.
	// @Dirty
	public static boolean IsUnitClause(SolveState s, double[] clause, double clauseSize, NumberReference literal) {
		boolean success;
		double i, l, v, unassigned, falsec, unassignc;

		success = false;

		falsec = 0d;
		unassignc = 0d;
		unassigned = -100000d; // Dummy. Guarateed to hold a value if unassignc

		for (i = 0d; i < clauseSize; i = i + 1d) {
			l = clause[(int) i];

			v = ComputeValue(s, l);
			if (v == 0d) {
				falsec = falsec + 1d;
			}

			if (v == -1d) {
				unassigned = l;
				unassignc = unassignc + 1d;
			}
		}

		if (unassignc == 1d && (falsec == (clauseSize - 1d) || (clauseSize == 1d))) {
			success = true;
			literal.numberValue = unassigned;
		}

		return success;
	}

	public static boolean AreArraysEqual(DynamicArrayNumbers arr1, double[] arr2) {
		boolean success, toBreak;
		double i;

		success = true;
		toBreak = false;

		if (arr1.length == arr2.length) {
			for (i = 0d; i < arr1.length && !toBreak; i = i + 1d) {
				if (arr1.array[(int) i] != arr2[(int) i]) {
					toBreak = true;
					success = false;
				}
			}
		} else {
			success = false;
		}

		return success;
	}

	public static void UpdateGraph(SolveState s, double var, DynamicArrayNumbers clause) {
		double i, lit;

		ImplicationNode node;

		node = s.nodes[(int) var];
		node.value = s.assignments[(int) var];
		node.level = s.level;

		// Update parents
		for (i = 0d; i < clause.length; i = i + 1d) {
			lit = abs(clause.array[(int) i]) - 1d;
			if (lit != var) {
				ArrayOfImplicationNodesAdd(node.parents, s.nodes[(int) lit]);
				ArrayOfImplicationNodesAdd(s.nodes[(int) lit].children, node);
			}
		}
		node.clause = clause;
		node.hasClause = true;
	}

	public static void UpdateGraphNoClause(SolveState s, double var) {
		ImplicationNode node;

		node = s.nodes[(int) var];
		node.value = s.assignments[(int) var];
		node.level = s.level;
	}

	// A unit clause has all of its literals but 1 assigned to 0. Then, the sole
	// unassigned literal must be assigned to value 1. Unit propagation is the
	// process of iteratively applying the unit clause rule.
	// :return: None if no conflict is detected, else return the literal
	public static boolean UnitPropagate(SolveState s, DynamicArrayNumbers result) {
		boolean success, toBreak;

		boolean isUnit, found;
		BooleanReference foundReference;

		double i, j, propVar, propLit;
		NumberReference r;

		ArrayOfArraysOfNumbers propClauses; // These two are paired
		DynamicArrayNumbers propUnitLits; // These two are paired

		DynamicArrayNumbers literalsCopy;
		double[] literals;
		double literalsSize; // See note above IsUnitClause

		DynamicArrayNumbers learnt;

		double cval;

		foundReference = CreateBooleanReference(false);

		toBreak = false;
		success = false;

		propClauses = CreateArrayOfArraysOfNumbers();
		propUnitLits = CreateDynamicArrayNumbersWithInitialCapacity(10d);

		for (; !toBreak;) {
			propClauses.size = 0d;
			propUnitLits.length = 0d;

			for (i = 0d; i < (s.clausesCount + s.learnts.size) && !toBreak; i = i + 1d) {
				// Iterate over the union of the initial clauses and the learnt ones
				if (i < s.clausesCount) {
					literals = s.clauses[(int) i].literals;
					literalsSize = literals.length;
				} else {
					learnt = s.learnts.data[(int) (i - s.clausesCount)];
					literals = learnt.array;
					literalsSize = learnt.length;
				}

				cval = ComputeClause(s, literals, literalsSize);

				if (cval == -1d) {
					r = new NumberReference();

					isUnit = IsUnitClause(s, literals, literalsSize, r);
					if (isUnit) {
						// Don't add duplicate clauses to propagate queue
						found = false;

						j = DynamicArrayNumbersIndexOf(propUnitLits, r.numberValue, foundReference);
						if (foundReference.booleanValue && AreArraysEqual(propClauses.data[(int) j], literals)) {
							found = true;
						}

						if (!found) {
							literalsCopy = CreateDynamicArrayNumbersWithInitialCapacity(10d);
							for (j = 0d; j < literalsSize; j = j + 1d) {
								DynamicArrayAddNumber(literalsCopy, literals[(int) j]);
							}
							ArrayOfArraysOfNumbersAdd(propClauses, literalsCopy);

							DynamicArrayAddNumber(propUnitLits, r.numberValue);
						}
					}
					delete(r);
				} else if (cval == 0d) {
					toBreak = true;
					success = true;
					result.array = literals;
					result.length = literalsSize;
				}
			}

			if (!toBreak) {
				if (propClauses.size == 0d) {
					toBreak = true;
					success = false;
				} else {
					for (i = 0d; i < propClauses.size; i = i + 1d) {
						propLit = propUnitLits.array[(int) i];
						propVar = abs(propLit);
						if (propLit > 0d) {
							s.assignments[(int) (propVar - 1d)] = 1d;
						} else {
							s.assignments[(int) (propVar - 1d)] = 0d;
						}

						UpdateGraph(s, propVar - 1d, propClauses.data[(int) i]);
						if (s.level != 0d) {
							DynamicArrayAddNumber(s.propagateHistory.data[(int) s.level], propLit);
						}
					}
				}
			}
		}

		delete(propClauses);
		delete(propUnitLits);

		return success;
	}

	// for a given variable x:
	// C(x,p) = # unresolved clauses in which x appears positively
	// C(x,n) = # unresolved clauses in which x appears negatively
	// find a variable a such that C(a,p) is max,
	// a variable b such that C(b,n) is max
	// if C(a,p) > C(b,n), assign a to 1, else assign b to 0
	public static double PickBranchingVariable(SolveState s, NumberReference value) {
		double[] vPos;
		double[] vNeg;

		double[] cl;

		double i, j;
		double l;

		double pc, nc, pi, ni;

		double result;

		vPos = new double[(int) s.varsCount];
		vNeg = new double[(int) s.varsCount];

		for (i = 0d; i < s.varsCount; i = i + 1d) {
			if (s.assignments[(int) i] == -1d) {
				vPos[(int) i] = 0d;
				vNeg[(int) i] = 0d;
			} else {
				vPos[(int) i] = -1d;
				vNeg[(int) i] = -1d;
			}
		}

		for (i = 0d; i < s.clausesCount; i = i + 1d) {
			cl = s.clauses[(int) i].literals;
			if (ComputeClause(s, cl, cl.length) == -1d) {
				// Clause is unresolved
				for (j = 0d; j < cl.length; j = j + 1d) {
					l = cl[(int) j];
					if (l > 0d) {
						if (vPos[(int) (l - 1d)] != -1d) {
							vPos[(int) (l - 1d)] = vPos[(int) (l - 1d)] + 1d;
						}
					} else {
						if (vNeg[(int) (abs(l) - 1d)] != -1d) {
							vNeg[(int) (abs(l) - 1d)] = vNeg[(int) (abs(l) - 1d)] + 1d;
						}
					}
				}
			}
		}

		pc = 0d;
		nc = 0d;
		pi = -1d;
		ni = -1d;

		for (i = 0d; i < s.varsCount; i = i + 1d) {
			if (vPos[(int) i] > pc) {
				pc = vPos[(int) i];
				pi = i;
			}
			if (vNeg[(int) i] > nc) {
				nc = vNeg[(int) i];
				ni = i;
			}
		}

		if (pc > nc) {
			result = pi;
			value.numberValue = 1d;
		} else {
			result = ni;
			value.numberValue = 0d;
		}

		delete(vPos);
		delete(vNeg);

		return result;
	}

	// Analyze the most recent conflict and learn a new clause from the conflict.
	// - Find the cut in the implication graph that led to the conflict
	// - Derive a new clause which is the negation of the assignments that led to
	// the conflict
	// Returns the level to backtrack to.
	public static double ConflictAnalyze(SolveState s, DynamicArrayNumbers confCl) {
		double l;
		double i, j;
		double lit, v;
		boolean found;

		double lastAssigned;

		DynamicArrayNumbers assignHistory;
		DynamicArrayNumbers propHistory;

		DynamicArrayNumbers poolLits;
		DynamicArrayNumbers doneVars;
		DynamicArrayNumbers prevLevelLits;
		DynamicArrayNumbers currLevelLits;

		ImplicationNode node;

		DynamicArrayNumbers learnt;

		boolean toBreak, toBreakLits;

		toBreak = false;

		lastAssigned = -1000d; // Dummy

		if (s.level == 0d) {
			l = -1d;
		} else {
			// Concat the latest branch with the propagate history

			assignHistory = CreateDynamicArrayNumbersWithInitialCapacity(10d);
			DynamicArrayAddNumber(assignHistory, s.branchingHistory.array[(int) s.level] + 1d);

			propHistory = s.propagateHistory.data[(int) s.level];

			for (i = 0d; i < propHistory.length; i = i + 1d) {
				DynamicArrayAddNumber(assignHistory, propHistory.array[(int) i]);
			}

			poolLits = CreateDynamicArrayNumbersWithInitialCapacity(10d);
			for (i = 0d; i < confCl.length; i = i + 1d) {
				found = DynamicArrayNumbersIsInArray(poolLits, confCl.array[(int) i]);
				if (!found) {
					DynamicArrayAddNumber(poolLits, confCl.array[(int) i]);
				}
			}

			doneVars = CreateDynamicArrayNumbersWithInitialCapacity(10d);
			prevLevelLits = CreateDynamicArrayNumbersWithInitialCapacity(10d);
			currLevelLits = CreateDynamicArrayNumbersWithInitialCapacity(10d);

			for (; !toBreak;) {
				for (i = 0d; i < poolLits.length; i = i + 1d) {
					lit = poolLits.array[(int) i];
					if (s.nodes[(int) (abs(lit) - 1d)].level == s.level) {
						found = DynamicArrayNumbersIsInArray(currLevelLits, lit);
						if (!found) {
							DynamicArrayAddNumber(currLevelLits, lit);
						}
					} else {
						found = DynamicArrayNumbersIsInArray(prevLevelLits, lit);
						if (!found) {
							DynamicArrayAddNumber(prevLevelLits, lit);
						}
					}
				}

				if (currLevelLits.length == 1d) {
					toBreak = true;
				} else {
					// According to the assign history, separate the latest assigned variable
					// with the rest in `clause`

					toBreakLits = false;

					j = -1000d; // Dummy

					for (i = assignHistory.length - 1d; i >= 0d && !toBreakLits; i = i - 1d) {
						lit = assignHistory.array[(int) i];

						for (j = 0d; j < currLevelLits.length && !toBreakLits; j = j + 1d) {
							v = currLevelLits.array[(int) j];
							if (v == lit || v == -lit) {
								// Found the last assigned literal
								lastAssigned = lit;
								toBreakLits = true;
							}
						}
					}

					DynamicArrayRemoveNumber(currLevelLits, j - 1d); // Remove the last assigned from the current level

					found = DynamicArrayNumbersIsInArray(doneVars, abs(lastAssigned));
					if (!found) {
						DynamicArrayAddNumber(doneVars, abs(lastAssigned));
					}

					poolLits.length = 0d;

					node = s.nodes[(int) (abs(lastAssigned) - 1d)];
					if (node.hasClause) {
						for (j = 0d; j < node.clause.length; j = j + 1d) {
							v = node.clause.array[(int) j];

							found = DynamicArrayNumbersIsInArray(doneVars, abs(v));
							if (!found) {
								DynamicArrayAddNumber(poolLits, v);
							}
						}
					}
				}
			}

			learnt = CreateDynamicArrayNumbersWithInitialCapacity(10d);
			for (i = 0d; i < currLevelLits.length; i = i + 1d) {
				DynamicArrayAddNumber(learnt, currLevelLits.array[(int) i]);
			}
			for (i = 0d; i < prevLevelLits.length; i = i + 1d) {
				DynamicArrayAddNumber(learnt, prevLevelLits.array[(int) i]);
			}
			ArrayOfArraysOfNumbersAdd(s.learnts, learnt);

			l = -100d; // Dummy

			if (prevLevelLits.length != 0d) {
				for (i = 0d; i < prevLevelLits.length; i = i + 1d) {
					l = max(l, s.nodes[(int) (abs(prevLevelLits.array[(int) i]) - 1d)].level);
				}
			} else {
				l = s.level - 1d;
			}

			delete(assignHistory.array);
			delete(assignHistory);

			delete(poolLits.array);
			delete(poolLits);
			delete(doneVars.array);
			delete(doneVars);
			delete(prevLevelLits.array);
			delete(prevLevelLits);

			delete(currLevelLits.array);
			delete(currLevelLits);
		}

		return l;
	}

	// Non-chronologically backtrack ("back jump") to the appropriate decision lvl,
	// where the first-assigned variable involved in the conflict was assigned
	public static void Backtrack(SolveState s, double l) {
		double i, j;

		ImplicationNode node;
		DynamicArrayNumbers hist;

		for (i = 0d; i < s.varsCount; i = i + 1d) {
			node = s.nodes[(int) i];

			if (node.level <= l) {
				// Remove any children that are higher level
				for (j = 0d; j < node.children.size; j = j + 1d) {
					if (node.children.data[(int) j].level > l) {
						ArrayOfImplicationNodesRemoveAt(node.children, j);
						j = j - 1d;
					}
				}
			} else {
				// Reset the node itself if it was a higher level
				node.value = -1d;
				node.level = -1d;
				node.parents.size = 0d;
				node.children.size = 0d;
				if (node.hasClause) {
					// We delete here because the clause was copied
					// in UnitPropagate. We are not deleting the
					// clauses we get in the Solve method nor the learnts ones.
					if (node.clause.length != 0d) {
						delete(node.clause.array);
					}
					delete(node.clause);
				}
				node.hasClause = false;

				s.assignments[(int) node.variableIndex] = -1;
			}
		}

		s.branchingVars.length = 0d;
		for (i = 0d; i < s.varsCount; i = i + 1d) {
			if ((s.assignments[(int) i] != -1d) && (s.nodes[(int) i].parents.size == 0d)) {
				DynamicArrayAddNumber(s.branchingVars, i);
			}
		}

		// Delete the extra histories to avoid leaks.
		// @Speed New ones are added when we branch (we could pool those).
		for (i = l + 1d; i < s.propagateHistory.size; i = i + 1d) {
			hist = s.propagateHistory.data[(int) i];
			if (hist.array.length != 0d) {
				delete(hist.array);
			}
			delete(hist);
		}
		s.propagateHistory.size = l + 1d;
		s.branchingHistory.length = l + 1d;
	}

	public static boolean Solve(ConjunctiveNormalForm cnf, double[] constrainedAssignments, BooleanArrayReference solution) {
		double i, l;
		boolean success, toBreak;

		Clause cl;

		SolveState s;

		boolean hasConfCl;
		DynamicArrayNumbers confCl; // This doesn't allocate. It's used as a return parameter.

		NumberReference btVal;
		double btVar;

		ImplicationNode node;

		s = new SolveState(); // @Leaks

		// We add the constrained assignments as just unit clauses.
		// e.g. the constraints [4, -5, -6] literally become the clauses [4], [-5], [-6]
		s.clauses = new Clause[(int) (cnf.clauses.length + constrainedAssignments.length)];
		for (i = 0d; i < cnf.clauses.length; i = i + 1d) {
			s.clauses[(int) i] = cnf.clauses[(int) i];
		}
		for (i = 0d; i < constrainedAssignments.length; i = i + 1d) {
			cl = new Clause();
			cl.literals = new double[1];
			cl.literals[0] = constrainedAssignments[(int) i];

			s.clauses[(int) (i + cnf.clauses.length)] = cl;
		}

		s.level = 0d;
		s.branchingCount = 0d;

		s.varsCount = GetNumberOfVariables(cnf);
		s.clausesCount = s.clauses.length;

		s.learnts = CreateArrayOfArraysOfNumbers();
		s.assignments = new double[(int) s.varsCount];

		for (i = 0d; i < s.varsCount; i = i + 1d) {
			s.assignments[(int) i] = -1d;
		}

		s.nodes = new ImplicationNode[(int) s.varsCount];
		for (i = 0d; i < s.varsCount; i = i + 1d) {
			node = new ImplicationNode();
			node.children = CreateArrayOfImplicationNodes();
			node.parents = CreateArrayOfImplicationNodes();
			node.variableIndex = i;
			s.nodes[(int) i] = node;
		}

		s.branchingVars = CreateDynamicArrayNumbersWithInitialCapacity(10d);

		s.branchingHistory = CreateDynamicArrayNumbersWithInitialCapacity(10d);
		s.propagateHistory = CreateArrayOfArraysOfNumbers();

		// Add dummies to histories in order for "s.level" to match as an index
		ArrayOfArraysOfNumbersAdd(s.propagateHistory, CreateDynamicArrayNumbersWithInitialCapacity(10d));
		DynamicArrayAddNumber(s.branchingHistory, 0d);

		confCl = CreateDynamicArrayNumbersWithInitialCapacity(10d);
		btVal = new NumberReference();

		// Initially set success to true.
		// If we fail we set it to false and break the loop.
		success = true;

		toBreak = false;

		for (; !AreAllVariablesAssigned(s) && !toBreak;) {
			hasConfCl = UnitPropagate(s, confCl);
			if (hasConfCl) {
				// There is conflict in unit propagation

				l = ConflictAnalyze(s, confCl);
				if (l < 0d) {
					success = false;
					toBreak = true;
				} else {
					Backtrack(s, l);
					s.level = l;
				}
			} else if (AreAllVariablesAssigned(s)) {
				toBreak = true;
				success = true;
			} else {
				// Branching
				s.level = s.level + 1d;
				s.branchingCount = s.branchingCount + 1d;

				btVar = PickBranchingVariable(s, btVal);

				DynamicArrayAddNumber(s.branchingVars, btVar);

				s.assignments[(int) btVar] = btVal.numberValue;
				ArrayOfArraysOfNumbersAdd(s.propagateHistory, CreateDynamicArrayNumbersWithInitialCapacity(10d));
				DynamicArrayAddNumber(s.branchingHistory, btVar);
				UpdateGraphNoClause(s, btVar);
			}
		}

		if (success) {
			solution.booleanArray = AssignmentsToBooleanArray(s.assignments);
		}

		for (i = 0d; i < s.clausesCount; i = i + 1d) {
			delete(s.clauses[(int) i]);
		}
		delete(s.clauses);
		delete(s.assignments);

		for (i = 0d; i < s.varsCount; i = i + 1d) {
			node = s.nodes[(int) i];
			delete(node.children);
			delete(node.parents);
			delete(node);
		}
		delete(s.nodes);

		for (i = 0d; i < s.learnts.size; i = i + 1d) {
			delete(s.learnts.data[(int) i].array);
			delete(s.learnts.data[(int) i]);
		}
		delete(s.learnts);

		for (i = 0d; i < s.propagateHistory.size; i = i + 1d) {
			delete(s.propagateHistory.data[(int) i].array);
			delete(s.propagateHistory.data[(int) i]);
		}
		delete(s.propagateHistory);

		delete(s.branchingVars.array);
		delete(s.branchingVars);

		delete(s.branchingHistory.array);
		delete(s.branchingHistory);

		delete(confCl);
		delete(btVal);

		delete(s);

		return success;
	}

	public static boolean IsValidConjunctiveNormalForm(ConjunctiveNormalForm cnf) {
		double totalVars, i, j, lit;
		double[] varsSet;
		Clause cl;

		boolean error;

		totalVars = GetNumberOfVariables(cnf);
		error = false;

		varsSet = new double[(int) totalVars];
		for (i = 0d; (i < cnf.clauses.length && !error); i = i + 1d) {
			cl = cnf.clauses[(int) i];

			for (j = 0d; (j < cl.literals.length && !error); j = j + 1d) {
				lit = abs(cl.literals[(int) j]);

				if (lit > 0d && lit <= totalVars) {
					varsSet[(int) (lit - 1d)] = 1d;
				} else {
					error = true; // Out of range
				}
			}
		}

		for (i = 0d; (i < totalVars && !error); i = i + 1d) {
			if (varsSet[(int) i] != 1d) {
				error = true;
			}
		}

		delete(varsSet);

		return !error;
	}

	public static boolean IsSatisfiable(ConjunctiveNormalForm cnf) {
		BooleanArrayReference solution;
		double[] constaints;
		boolean success;

		if (!IsValidConjunctiveNormalForm(cnf)) {
			success = false;
		} else {
			solution = new BooleanArrayReference(); // Discarded
			constaints = new double[0];

			success = Solve(cnf, constaints, solution);

			delete(solution);
			delete(constaints);
		}

		return success;
	}

	public static boolean IsSatisfiableWithAssignments(ConjunctiveNormalForm cnf, double[] assignments) {
		BooleanArrayReference solution;
		boolean success;

		if (!IsValidConjunctiveNormalForm(cnf)) {
			success = false;
		} else {
			solution = new BooleanArrayReference(); // Discarded
			success = Solve(cnf, assignments, solution);
		}
		return success;
	}

	public static boolean FindSolution(ConjunctiveNormalForm cnf, BooleanArrayReference solution) {
		double[] constaints;
		boolean success;

		if (!IsValidConjunctiveNormalForm(cnf)) {
			success = false;
		} else {
			constaints = new double[0];
			success = Solve(cnf, constaints, solution);
			delete(constaints);
		}
		return success;
	}

	public static boolean FindSolutionWithAssignments(ConjunctiveNormalForm cnf, BooleanArrayReference solution,
			double[] assignments) {
		boolean success;
		if (!IsValidConjunctiveNormalForm(cnf)) {
			success = false;
		} else {
			success = Solve(cnf, assignments, solution);
		}
		return success;
	}

	public static void delete(Object object) {
		// Java has garbage collection.
	}
}
