package tests;

import BooleanFormula.BooleanFormula.*;
import JSON.structures.Element;
import JSON.structures.ElementReference;
import references.references.*;

import static BooleanFormula.BooleanFormula.BooleanFormula.*;
import static JSON.parser.parser.ReadJSON;
import static arrays.arrays.arrays.StringsEqual;
import static java.lang.Math.*;
import static references.references.references.*;
import static testing.testing.testing.*;

// Benchmark: 531

public class tests {
    public static double test(){
        NumberReference failures;

        failures = CreateNumberReference(0d);

        testIsSatisfiable(failures);
        testFindSolution(failures);

        return failures.numberValue;
    }

    public static void testFindSolution(NumberReference failures) {
        ConjunctiveNormalForm cnf;
        BooleanArrayReference solution;

        cnf = new ConjunctiveNormalForm();
        solution = new BooleanArrayReference();

        CNFExample1(cnf);

        AssertTrue(IsValidConjunctiveNormalForm(cnf), failures);

        solution.booleanArray = new boolean [6];
        solution.booleanArray[0] = true;
        solution.booleanArray[1] = false;
        solution.booleanArray[2] = false;
        solution.booleanArray[3] = false;
        solution.booleanArray[4] = true;
        solution.booleanArray[5] = true;
        AssertTrue(Evaluate(cnf, solution.booleanArray), failures);

        solution = new BooleanArrayReference();
        CNFExample2(cnf);
        AssertFalse(FindSolution(cnf, solution), failures);

        solution = new BooleanArrayReference();
        CNFExample3(cnf);
        AssertTrue(FindSolution(cnf, solution), failures);
        AssertTrue(Evaluate(cnf, solution.booleanArray), failures);

        solution = new BooleanArrayReference();
        CNFExample5(cnf);
        AssertFalse(FindSolution(cnf, solution), failures);

        solution = new BooleanArrayReference();
        CNFExample6(cnf);
        AssertFalse(FindSolution(cnf, solution), failures);

        solution = new BooleanArrayReference();
        CNFExample7(cnf);
        AssertTrue(FindSolution(cnf, solution), failures);
        AssertTrue(Evaluate(cnf, solution.booleanArray), failures);

        solution = new BooleanArrayReference();
        CNFExample8(cnf);
        AssertTrue(FindSolution(cnf, solution), failures);
        AssertTrue(Evaluate(cnf, solution.booleanArray), failures);
    }

    public static void testIsSatisfiable(NumberReference failures) {
        ConjunctiveNormalForm cnf;

        cnf = new ConjunctiveNormalForm();

        CNFExample1(cnf);

        AssertTrue(IsValidConjunctiveNormalForm(cnf), failures);
        AssertTrue(IsSatisfiable(cnf), failures);

        // aim-100-1_6-no-1.cnf - Not Satisfiable
        AssertTrue(CNFExample2(cnf), failures);
        AssertFalse(IsSatisfiable(cnf), failures);

        // aim-50-1_6-yes1-4.cnf - Satisfiable
        AssertTrue(CNFExample3(cnf), failures);
        AssertTrue(IsSatisfiable(cnf), failures);

        // bf0432-007.cnf - ?
        //AssertTrue(CNFExample4(cnf), failures);
        //AssertFalse(IsSatisfiable(cnf), failures);

        // dubois20.cnf - Not satisfiable
        AssertTrue(CNFExample5(cnf), failures);
        AssertFalse(IsSatisfiable(cnf), failures);

        // dubois21.cnf - Not satisfiable
        AssertTrue(CNFExample6(cnf), failures);
        AssertFalse(IsSatisfiable(cnf), failures);

        //
        AssertTrue(CNFExample7(cnf), failures);
        AssertTrue(IsSatisfiable(cnf), failures);

        AssertTrue(CNFExample8(cnf), failures);
        AssertTrue(IsSatisfiable(cnf), failures);
    }

    public static void CNFExample1(ConjunctiveNormalForm cnf) {
        JSONToCNF(cnf, "[[1, -2, -3], [-4, 5, 6]]".toCharArray());
    }

    public static boolean CNFExample7(ConjunctiveNormalForm cnf) {
        return JSONToCNF(cnf, "[[-2,1],[-3,-2],[-3,-2,-1],[3,2,-1],[-3,2,1],[3,-2,1],[-4,2],[-5,-4],[-5,-4,-2],[5,4,-2],[-5,4,2],[5,-4,2],[-6,4],[-7,-6],[-7,-6,-4],[7,6,-4],[-7,6,4],[7,-6,4],[-8,6],[-9,-8],[-9,-8,-6],[9,8,-6],[-9,8,6],[9,-8,6],[-10,8],[-11,-10],[-11,-10,-8],[11,10,-8],[-11,10,8],[11,-10,8],[-12,10],[-13,-12],[-13,-12,-10],[13,12,-10],[-13,12,10],[13,-12,10],[-14,12],[-15,-14],[-15,-14,-12],[15,14,-12],[-15,14,12],[15,-14,12],[-16,14],[-17,-16],[-17,-16,-14],[17,16,-14],[-17,16,14],[17,-16,14],[-18,16],[-19,-18],[-19,-18,-16],[19,18,-16],[-19,18,16],[19,-18,16],[-20,18],[-21,-20],[-21,-20,-18],[21,20,-18],[-21,20,18],[21,-20,18],[-22,20],[-23,-22],[-23,-22,-20],[23,22,-20],[-23,22,20],[23,-22,20],[-24,22],[-25,-24],[-25,-24,-22],[25,24,-22],[-25,24,22],[25,-24,22],[-26,24],[-27,-26],[-27,-26,-24],[27,26,-24],[-27,26,24],[27,-26,24],[-28,26],[-29,-28],[-29,-28,-26],[29,28,-26],[-29,28,26],[29,-28,26],[28,-30],[-31,-30],[-31,-28,-30],[31,-28,30],[-31,28,30],[31,28,-30],[-33,-32,-3],[33,32,-3],[-33,32,3],[33,-32,3],[-35,-34,-32],[35,34,-32],[-35,34,32],[35,-34,32],[-37,-34,36],[37,-34,-36],[-37,34,-36],[37,34,36],[-39,-38,-5],[39,38,-5],[-39,38,5],[39,-38,5],[-35,-40,-38],[35,40,-38],[-35,40,38],[35,-40,38],[-42,-41,-40],[42,41,-40],[-42,41,40],[42,-41,40],[-36,-41,43],[36,-41,-43],[-36,41,-43],[36,41,43],[-44,-7,29],[44,-7,-29],[44,7,29],[-44,7,-29],[-33,-45,-44],[33,45,-44],[-33,45,44],[33,-45,44],[-37,-36,-45],[37,36,-45],[-37,36,45],[37,-36,45],[-37,-46,-9],[37,46,-9],[-37,46,9],[37,-46,9],[-36,-43,-46],[36,43,-46],[-36,43,46],[36,-43,46],[-39,-47,-11],[39,47,-11],[-39,47,11],[39,-47,11],[-33,-48,-47],[33,48,-47],[-33,48,47],[33,-48,47],[-37,-36,-48],[37,36,-48],[-37,36,48],[37,-36,48],[-39,-49,-13],[39,49,-13],[-39,49,13],[39,-49,13],[-33,-36,-49],[33,36,-49],[-33,36,49],[33,-36,49],[-50,-15,29],[50,-15,-29],[50,15,29],[-50,15,-29],[-35,-37,-50],[35,37,-50],[-35,37,50],[35,-37,50],[-39,-35,-17],[39,35,-17],[-39,35,17],[39,-35,17],[-39,-51,-19],[39,51,-19],[-39,51,19],[39,-51,19],[-35,-52,-51],[35,52,-51],[-35,52,51],[35,-52,51],[-37,-52,42],[37,-52,-42],[-37,52,-42],[37,52,42],[-53,-21,29],[53,-21,-29],[53,21,29],[-53,21,-29],[-33,-54,-53],[33,54,-53],[-33,54,53],[33,-54,53],[-35,-54,42],[35,-54,-42],[-35,54,-42],[35,54,42],[-33,-23,42],[33,-23,-42],[-33,23,-42],[33,23,42],[-55,-25,29],[55,-25,-29],[55,25,29],[-55,25,-29],[-33,-56,-55],[33,56,-55],[-33,56,55],[33,-56,55],[-35,-56,36],[35,-56,-36],[-35,56,-36],[35,56,36],[-39,-57,-27],[39,57,-27],[-39,57,27],[39,-57,27],[-58,-57,29],[58,-57,-29],[58,57,29],[-58,57,-29],[-35,-59,-58],[35,59,-58],[-35,59,58],[35,-59,58],[-37,-59,-36],[37,-59,36],[-37,59,36],[37,59,-36],[-37,-60,-31],[37,60,-31],[-37,60,31],[37,-60,31],[-42,-61,-60],[42,61,-60],[-42,61,60],[42,-61,60],[-36,-61,43],[36,-61,-43],[-36,61,-43],[36,61,43],[-39,-62,-30],[39,62,-30],[-39,62,30],[39,-62,30],[-33,-63,-62],[33,63,-62],[-33,63,62],[33,-63,62],[-42,-64,-63],[42,64,-63],[-42,64,63],[42,-64,63],[-36,-64,-43],[36,-64,43],[-36,64,43],[36,64,-43]]".toCharArray());
    }

    public static boolean CNFExample6(ConjunctiveNormalForm cnf) {
        return JSONToCNF(cnf, "[[41,42,1],[-41,-42,1],[41,-42,-1],[-41,42,-1],[1,43,2],[-1,-43,2],[1,-43,-2],[-1,43,-2],[2,44,3],[-2,-44,3],[2,-44,-3],[-2,44,-3],[3,45,4],[-3,-45,4],[3,-45,-4],[-3,45,-4],[4,46,5],[-4,-46,5],[4,-46,-5],[-4,46,-5],[5,47,6],[-5,-47,6],[5,-47,-6],[-5,47,-6],[6,48,7],[-6,-48,7],[6,-48,-7],[-6,48,-7],[7,49,8],[-7,-49,8],[7,-49,-8],[-7,49,-8],[8,50,9],[-8,-50,9],[8,-50,-9],[-8,50,-9],[9,51,10],[-9,-51,10],[9,-51,-10],[-9,51,-10],[10,52,11],[-10,-52,11],[10,-52,-11],[-10,52,-11],[11,53,12],[-11,-53,12],[11,-53,-12],[-11,53,-12],[12,54,13],[-12,-54,13],[12,-54,-13],[-12,54,-13],[13,55,14],[-13,-55,14],[13,-55,-14],[-13,55,-14],[14,56,15],[-14,-56,15],[14,-56,-15],[-14,56,-15],[15,57,16],[-15,-57,16],[15,-57,-16],[-15,57,-16],[16,58,17],[-16,-58,17],[16,-58,-17],[-16,58,-17],[17,59,18],[-17,-59,18],[17,-59,-18],[-17,59,-18],[18,60,19],[-18,-60,19],[18,-60,-19],[-18,60,-19],[19,61,20],[-19,-61,20],[19,-61,-20],[-19,61,-20],[20,62,63],[-20,-62,63],[20,-62,-63],[-20,62,-63],[21,62,63],[-21,-62,63],[21,-62,-63],[-21,62,-63],[22,61,21],[-22,-61,21],[22,-61,-21],[-22,61,-21],[23,60,22],[-23,-60,22],[23,-60,-22],[-23,60,-22],[24,59,23],[-24,-59,23],[24,-59,-23],[-24,59,-23],[25,58,24],[-25,-58,24],[25,-58,-24],[-25,58,-24],[26,57,25],[-26,-57,25],[26,-57,-25],[-26,57,-25],[27,56,26],[-27,-56,26],[27,-56,-26],[-27,56,-26],[28,55,27],[-28,-55,27],[28,-55,-27],[-28,55,-27],[29,54,28],[-29,-54,28],[29,-54,-28],[-29,54,-28],[30,53,29],[-30,-53,29],[30,-53,-29],[-30,53,-29],[31,52,30],[-31,-52,30],[31,-52,-30],[-31,52,-30],[32,51,31],[-32,-51,31],[32,-51,-31],[-32,51,-31],[33,50,32],[-33,-50,32],[33,-50,-32],[-33,50,-32],[34,49,33],[-34,-49,33],[34,-49,-33],[-34,49,-33],[35,48,34],[-35,-48,34],[35,-48,-34],[-35,48,-34],[36,47,35],[-36,-47,35],[36,-47,-35],[-36,47,-35],[37,46,36],[-37,-46,36],[37,-46,-36],[-37,46,-36],[38,45,37],[-38,-45,37],[38,-45,-37],[-38,45,-37],[39,44,38],[-39,-44,38],[39,-44,-38],[-39,44,-38],[40,43,39],[-40,-43,39],[40,-43,-39],[-40,43,-39],[41,42,-40],[-41,-42,-40],[41,-42,40],[-41,42,40]]".toCharArray());
    }

    public static boolean CNFExample5(ConjunctiveNormalForm cnf) {
        return JSONToCNF(cnf, "[[39,40,1],[-39,-40,1],[39,-40,-1],[-39,40,-1],[1,41,2],[-1,-41,2],[1,-41,-2],[-1,41,-2],[2,42,3],[-2,-42,3],[2,-42,-3],[-2,42,-3],[3,43,4],[-3,-43,4],[3,-43,-4],[-3,43,-4],[4,44,5],[-4,-44,5],[4,-44,-5],[-4,44,-5],[5,45,6],[-5,-45,6],[5,-45,-6],[-5,45,-6],[6,46,7],[-6,-46,7],[6,-46,-7],[-6,46,-7],[7,47,8],[-7,-47,8],[7,-47,-8],[-7,47,-8],[8,48,9],[-8,-48,9],[8,-48,-9],[-8,48,-9],[9,49,10],[-9,-49,10],[9,-49,-10],[-9,49,-10],[10,50,11],[-10,-50,11],[10,-50,-11],[-10,50,-11],[11,51,12],[-11,-51,12],[11,-51,-12],[-11,51,-12],[12,52,13],[-12,-52,13],[12,-52,-13],[-12,52,-13],[13,53,14],[-13,-53,14],[13,-53,-14],[-13,53,-14],[14,54,15],[-14,-54,15],[14,-54,-15],[-14,54,-15],[15,55,16],[-15,-55,16],[15,-55,-16],[-15,55,-16],[16,56,17],[-16,-56,17],[16,-56,-17],[-16,56,-17],[17,57,18],[-17,-57,18],[17,-57,-18],[-17,57,-18],[18,58,19],[-18,-58,19],[18,-58,-19],[-18,58,-19],[19,59,60],[-19,-59,60],[19,-59,-60],[-19,59,-60],[20,59,60],[-20,-59,60],[20,-59,-60],[-20,59,-60],[21,58,20],[-21,-58,20],[21,-58,-20],[-21,58,-20],[22,57,21],[-22,-57,21],[22,-57,-21],[-22,57,-21],[23,56,22],[-23,-56,22],[23,-56,-22],[-23,56,-22],[24,55,23],[-24,-55,23],[24,-55,-23],[-24,55,-23],[25,54,24],[-25,-54,24],[25,-54,-24],[-25,54,-24],[26,53,25],[-26,-53,25],[26,-53,-25],[-26,53,-25],[27,52,26],[-27,-52,26],[27,-52,-26],[-27,52,-26],[28,51,27],[-28,-51,27],[28,-51,-27],[-28,51,-27],[29,50,28],[-29,-50,28],[29,-50,-28],[-29,50,-28],[30,49,29],[-30,-49,29],[30,-49,-29],[-30,49,-29],[31,48,30],[-31,-48,30],[31,-48,-30],[-31,48,-30],[32,47,31],[-32,-47,31],[32,-47,-31],[-32,47,-31],[33,46,32],[-33,-46,32],[33,-46,-32],[-33,46,-32],[34,45,33],[-34,-45,33],[34,-45,-33],[-34,45,-33],[35,44,34],[-35,-44,34],[35,-44,-34],[-35,44,-34],[36,43,35],[-36,-43,35],[36,-43,-35],[-36,43,-35],[37,42,36],[-37,-42,36],[37,-42,-36],[-37,42,-36],[38,41,37],[-38,-41,37],[38,-41,-37],[-38,41,-37],[39,40,-38],[-39,-40,-38],[39,-40,38],[-39,40,38]]".toCharArray());
    }

    public static boolean CNFExample4(ConjunctiveNormalForm cnf) {
        return JSONToCNF(cnf, "[[700],[-985],[1039],[700],[-701],[699],[663,-664],[-663,664],[699,701],[-699,-701],[-667,-706],[-665,-706],[-669,-706],[699,663],[-699,-663],[985,-701],[-985,701],[985,-664],[-985,664],[701,-664],[-701,664],[-39,-988],[37,-988],[39,-989],[-37,-989],[-989,987],[-988,987],[201,37],[-201,-37],[-202,38],[203,39],[-203,-39],[207,201],[-207,-201],[-208,202],[209,203],[-209,-203],[62,-204],[-62,204],[62,-207],[-62,207],[62,-214],[-62,214],[62,-259],[-62,259],[62,-288],[-62,288],[62,-436],[-62,436],[62,-495],[-62,495],[62,-544],[-62,544],[62,-597],[-62,597],[62,-639],[-62,639],[62,-708],[-62,708],[62,-736],[-62,736],[62,-762],[-62,762],[62,-778],[-62,778],[62,-827],[-62,827],[62,-838],[-62,838],[62,-854],[-62,854],[62,-914],[-62,914],[62,-949],[-62,949],[63,-205],[63,-208],[63,-215],[63,-260],[63,-289],[63,-437],[63,-496],[63,-545],[63,-598],[63,-640],[63,-709],[63,-737],[63,-763],[63,-779],[63,-828],[63,-839],[63,-855],[63,-915],[63,-950],[64,-206],[-64,206],[64,-209],[-64,209],[64,-216],[-64,216],[64,-261],[-64,261],[64,-290],[-64,290],[64,-438],[-64,438],[64,-497],[-64,497],[64,-546],[-64,546],[64,-599],[-64,599],[64,-641],[-64,641],[64,-710],[-64,710],[64,-738],[-64,738],[64,-764],[-64,764],[64,-780],[-64,780],[64,-829],[-64,829],[64,-840],[-64,840],[64,-856],[-64,856],[64,-916],[-64,916],[64,-951],[-64,951],[699,62],[663,62],[65,62],[685,62],[701,64],[663,64],[65,64],[685,64],[836,685],[-836,-685],[686,-836],[-686,836],[686,-837],[-686,837],[-884,-686],[-885,-686],[690,-697],[-690,697],[690,-698],[-690,698],[690,-730],[-690,730],[690,-885],[-690,885],[26,690],[-26,-690],[7,-881],[-7,881],[7,-884],[-7,884],[-286,-65],[-68,-65],[61,-67],[-61,67],[61,-68],[-61,68],[-547,-61],[-533,-61],[514,-533],[-514,533],[514,-534],[-514,534],[536,514],[-536,-514],[32,-536],[-32,536],[32,-537],[-32,537],[32,-568],[-32,568],[5,-547],[-5,547],[5,-548],[-5,548],[66,-286],[-66,286],[66,-287],[-66,287],[-857,-66],[-850,-66],[841,-848],[-841,848],[841,-849],[-841,849],[841,-850],[-841,850],[841,-878],[-841,878],[9,841],[-9,-841],[21,-857],[-21,857],[21,-858],[-21,858],[-667,-663],[-665,-663],[-669,-663],[494,-668],[-494,668],[494,-669],[-494,669],[-512,-494],[-513,-494],[470,-490],[-470,490],[470,-493],[-470,493],[470,-508],[-470,508],[470,-513],[-470,513],[34,470],[-34,-470],[25,-509],[-25,509],[25,-512],[-25,512],[796,665],[781,665],[797,665],[947,797],[-947,-797],[926,-947],[-926,947],[926,-948],[-926,948],[-945,-926],[-946,-926],[927,-938],[-927,938],[927,-941],[-927,941],[927,-946],[-927,946],[927,-962],[-927,962],[19,927],[-19,-927],[17,-942],[-17,942],[17,-945],[-17,945],[783,781],[-783,-781],[777,-782],[-777,782],[777,-783],[-777,783],[-924,-777],[-925,-777],[587,-917],[-587,917],[587,-918],[-587,918],[587,-925],[-587,925],[910,587],[-910,-587],[35,-909],[-35,909],[35,-910],[-35,910],[30,-921],[-30,921],[30,-924],[-30,924],[799,796],[-799,-796],[213,-798],[-213,798],[213,-799],[-213,799],[-775,-213],[-776,-213],[220,-233],[-220,233],[220,-234],[-220,234],[220,-424],[-220,424],[220,-776],[-220,776],[28,220],[-28,-220],[4,-772],[-4,772],[4,-775],[-4,775],[591,-666],[-591,666],[591,-667],[-591,667],[-600,-591],[-593,-591],[580,-592],[-580,592],[580,-593],[-580,593],[627,580],[-627,-580],[6,-627],[-6,627],[6,-628],[-6,628],[6,-638],[-6,638],[10,-600],[-10,600],[10,-601],[-10,601],[706,699],[-706,-699],[702,-706],[-702,706],[702,-707],[-702,707],[-889,-702],[-890,-702],[711,-718],[-711,718],[711,-719],[-711,719],[711,-756],[-711,756],[711,-890],[-711,890],[27,711],[-27,-711],[11,-886],[-11,886],[11,-889],[-11,889],[-42,-995],[40,-995],[42,-996],[-40,-996],[-996,994],[-995,994],[657,40],[-657,-40],[-658,41],[659,42],[-659,-42],[195,-657],[-195,657],[195,-660],[-195,660],[195,-676],[-195,676],[196,-658],[196,-661],[196,-677],[197,-659],[-197,659],[197,-662],[-197,662],[197,-678],[-197,678],[250,195],[-250,-195],[-251,196],[252,197],[-252,-197],[87,-235],[-87,235],[87,-238],[-87,238],[87,-250],[-87,250],[87,-262],[-87,262],[87,-271],[-87,271],[87,-309],[-87,309],[87,-315],[-87,315],[87,-324],[-87,324],[87,-339],[-87,339],[87,-439],[-87,439],[87,-521],[-87,521],[87,-551],[-87,551],[87,-604],[-87,604],[87,-620],[-87,620],[87,-739],[-87,739],[87,-765],[-87,765],[87,-793],[-87,793],[87,-824],[-87,824],[87,-830],[-87,830],[87,-861],[-87,861],[87,-901],[-87,901],[88,-236],[88,-239],[88,-251],[88,-263],[88,-272],[88,-310],[88,-316],[88,-325],[88,-340],[88,-440],[88,-522],[88,-552],[88,-605],[88,-621],[88,-740],[88,-766],[88,-794],[88,-825],[88,-831],[88,-862],[88,-902],[89,-237],[-89,237],[89,-240],[-89,240],[89,-252],[-89,252],[89,-264],[-89,264],[89,-273],[-89,273],[89,-311],[-89,311],[89,-317],[-89,317],[89,-326],[-89,326],[89,-341],[-89,341],[89,-441],[-89,441],[89,-523],[-89,523],[89,-553],[-89,553],[89,-606],[-89,606],[89,-622],[-89,622],[89,-741],[-89,741],[89,-767],[-89,767],[89,-795],[-89,795],[89,-826],[-89,826],[89,-832],[-89,832],[89,-863],[-89,863],[89,-903],[-89,903],[108,87],[651,87],[446,87],[464,87],[-109,88],[-652,88],[-447,88],[-465,88],[110,89],[653,89],[448,89],[466,89],[480,464],[-480,-464],[-481,465],[482,466],[-482,-466],[268,-477],[-268,477],[268,-480],[-268,480],[269,-478],[269,-481],[270,-479],[-270,479],[270,-482],[-270,482],[-498,-268],[-489,-268],[-493,-268],[-499,269],[-499,-493],[-499,-489],[-500,-270],[-489,-270],[-493,-270],[12,-486],[-12,486],[12,-489],[-12,489],[483,-498],[-483,498],[483,-501],[-483,501],[484,-499],[484,-502],[485,-500],[-485,500],[485,-503],[-485,503],[-668,-975],[495,-975],[668,-976],[-495,-976],[-976,483],[-975,483],[-496,484],[-668,-1001],[497,-1001],[668,-1002],[-497,-1002],[-1002,485],[-1001,485],[-458,-446],[-787,-446],[-679,-446],[-459,447],[-788,447],[-680,447],[-460,-448],[-789,-448],[-681,-448],[452,-679],[-452,679],[452,-682],[-452,682],[452,-821],[-452,821],[453,-680],[453,-683],[453,-822],[454,-681],[-454,681],[454,-684],[-454,684],[454,-823],[-454,823],[-952,-452],[-937,-452],[-941,-452],[-953,453],[-953,-941],[-953,-937],[-954,-454],[-937,-454],[-941,-454],[18,-934],[-18,934],[18,-937],[-18,937],[931,-952],[-931,952],[931,-955],[-931,955],[932,-953],[932,-956],[933,-954],[-933,954],[933,-957],[-933,957],[-948,-963],[949,-963],[948,-964],[-949,-964],[-964,931],[-963,931],[-950,932],[-948,-1003],[951,-1003],[948,-1004],[-951,-1004],[-1004,933],[-1003,933],[449,-787],[-449,787],[449,-790],[-449,790],[450,-788],[450,-791],[451,-789],[-451,789],[451,-792],[-451,792],[-806,-449],[-897,-449],[-917,-449],[-807,450],[-807,-917],[-807,-897],[-808,-451],[-897,-451],[-917,-451],[22,-897],[-22,897],[22,-898],[-22,898],[584,-806],[-584,806],[584,-809],[-584,809],[585,-807],[585,-810],[586,-808],[-586,808],[586,-811],[-586,811],[-782,-969],[778,-969],[782,-970],[-778,-970],[-970,584],[-969,584],[-779,585],[-782,-1005],[780,-1005],[782,-1006],[-780,-1006],[-1006,586],[-1005,586],[217,-455],[-217,455],[217,-458],[-217,458],[218,-456],[218,-459],[219,-457],[-219,457],[219,-460],[-219,460],[-227,-217],[-442,-217],[-233,-217],[-228,218],[-228,-233],[-228,-442],[-229,-219],[-442,-219],[-233,-219],[23,-442],[-23,442],[23,-443],[-23,443],[210,-227],[-210,227],[210,-230],[-210,230],[211,-228],[211,-231],[212,-229],[-212,229],[212,-232],[-212,232],[-798,-981],[214,-981],[798,-982],[-214,-982],[-982,210],[-981,210],[-215,211],[-798,-1007],[216,-1007],[798,-1008],[-216,-1008],[-1008,212],[-1007,212],[461,-651],[-461,651],[461,-654],[-461,654],[462,-652],[462,-655],[463,-653],[-463,653],[463,-656],[-463,656],[614,461],[-614,-461],[-615,462],[616,463],[-616,-463],[581,-614],[-581,614],[581,-617],[-581,617],[582,-615],[582,-618],[583,-616],[-583,616],[583,-619],[-583,619],[-645,-581],[-607,-581],[-592,-581],[-646,582],[-646,-592],[-646,-607],[-647,-583],[-607,-583],[-592,-583],[33,-607],[-33,607],[33,-608],[-33,608],[588,-645],[-588,645],[588,-648],[-588,648],[589,-646],[589,-649],[590,-647],[-590,647],[590,-650],[-590,650],[-666,-973],[639,-973],[666,-974],[-639,-974],[-974,588],[-973,588],[-640,589],[-666,-1009],[641,-1009],[666,-1010],[-641,-1010],[-1010,590],[-1009,590],[-123,-108],[-318,-108],[-303,-108],[-333,-108],[-124,109],[-319,109],[-304,109],[-334,109],[-125,-110],[-320,-110],[-305,-110],[-335,-110],[117,-333],[-117,333],[117,-336],[-117,336],[117,-518],[-117,518],[118,-334],[118,-337],[118,-519],[119,-335],[-119,335],[119,-338],[-119,338],[119,-520],[-119,520],[-530,-117],[-554,-117],[-534,-117],[-531,118],[-531,-534],[-531,-554],[-532,-119],[-554,-119],[-534,-119],[29,-554],[-29,554],[29,-555],[-29,555],[58,-527],[-58,527],[58,-530],[-58,530],[59,-528],[59,-531],[60,-529],[-60,529],[60,-532],[-60,532],[-67,-983],[204,-983],[67,-984],[-204,-984],[-984,58],[-983,58],[-205,59],[-67,-1011],[206,-1011],[67,-1012],[-206,-1012],[-1012,60],[-1011,60],[114,-303],[-114,303],[114,-306],[-114,306],[114,-312],[-114,312],[115,-304],[115,-307],[115,-313],[116,-305],[-116,305],[116,-308],[-116,308],[116,-314],[-116,314],[-694,-114],[-742,-114],[-698,-114],[-695,115],[-695,-698],[-695,-742],[-696,-116],[-742,-116],[-698,-116],[31,-742],[-31,742],[31,-743],[-31,743],[687,-691],[-687,691],[687,-694],[-687,694],[688,-692],[688,-695],[689,-693],[-689,693],[689,-696],[-689,696],[-837,-965],[838,-965],[837,-966],[-838,-966],[-966,687],[-965,687],[-839,688],[-837,-1013],[840,-1013],[837,-1014],[-840,-1014],[-1014,689],[-1013,689],[111,-318],[-111,318],[111,-321],[-111,321],[112,-319],[112,-322],[113,-320],[-113,320],[113,-323],[-113,323],[-842,-111],[-864,-111],[-848,-111],[-843,112],[-843,-848],[-843,-864],[-844,-113],[-864,-113],[-848,-113],[20,-864],[-20,864],[20,-865],[-20,865],[283,-842],[-283,842],[283,-845],[-283,845],[284,-843],[284,-846],[285,-844],[-285,844],[285,-847],[-285,847],[-287,-977],[288,-977],[287,-978],[-288,-978],[-978,283],[-977,283],[-289,284],[-287,-1015],[290,-1015],[287,-1016],[-290,-1016],[-1016,285],[-1015,285],[84,-120],[-84,120],[84,-123],[-84,123],[85,-121],[85,-124],[86,-122],[-86,122],[86,-125],[-86,125],[-712,-84],[-768,-84],[-718,-84],[-713,85],[-713,-718],[-713,-768],[-714,-86],[-768,-86],[-718,-86],[15,-768],[-15,768],[15,-769],[-15,769],[703,-712],[-703,712],[703,-715],[-703,715],[704,-713],[704,-716],[705,-714],[-705,714],[705,-717],[-705,717],[-707,-971],[708,-971],[707,-972],[-708,-972],[-972,703],[-971,703],[-709,704],[-707,-1017],[710,-1017],[707,-1018],[-710,-1018],[-1018,705],[-1017,705],[-45,-1020],[43,-1020],[45,-1021],[-43,-1021],[-1021,1019],[-1020,1019],[558,43],[-558,-43],[-559,44],[560,45],[-560,-45],[247,-558],[-247,558],[247,-561],[-247,561],[247,-624],[-247,624],[248,-559],[248,-562],[248,-625],[249,-560],[-249,560],[249,-563],[-249,563],[249,-626],[-249,626],[577,247],[-577,-247],[-578,248],[579,249],[-579,-249],[102,-471],[-102,471],[102,-474],[-102,474],[102,-577],[-102,577],[102,-723],[-102,723],[102,-749],[-102,749],[102,-871],[-102,871],[102,-894],[-102,894],[102,-928],[-102,928],[103,-472],[103,-475],[103,-578],[103,-724],[103,-750],[103,-872],[103,-895],[103,-929],[104,-473],[-104,473],[104,-476],[-104,476],[104,-579],[-104,579],[104,-725],[-104,725],[104,-751],[-104,751],[104,-873],[-104,873],[104,-896],[-104,896],[104,-930],[-104,930],[75,102],[105,102],[90,102],[-76,103],[-106,103],[-91,103],[77,104],[107,104],[92,104],[-99,-90],[-94,91],[-97,91],[-100,91],[-101,-92],[-527,-99],[-524,-99],[-516,100],[-328,100],[-528,100],[-528,-524],[-328,-524],[-516,-524],[-529,-101],[-524,-101],[564,524],[537,524],[535,-564],[-535,564],[535,-565],[-535,565],[14,535],[-14,-535],[330,327],[-330,-327],[-331,328],[332,329],[-332,-329],[-336,-330],[-339,-330],[-337,331],[-340,331],[-338,-332],[-341,-332],[518,515],[521,515],[-519,516],[-522,516],[520,517],[523,517],[-845,-96],[-874,-96],[-849,-96],[-846,97],[-846,-849],[-846,-874],[-847,-98],[-874,-98],[-849,-98],[8,-874],[-8,874],[8,-875],[-8,875],[-322,94],[-325,94],[-241,-105],[-221,-105],[-244,-105],[-242,106],[-222,106],[-245,106],[-243,-107],[-223,-107],[-246,-107],[803,244],[-785,245],[-810,245],[-804,245],[-810,-800],[-785,-800],[805,246],[833,803],[-671,804],[-819,804],[-834,804],[835,805],[-955,-833],[-961,-833],[-962,-833],[-956,834],[-956,-962],[-956,-961],[-957,-835],[-961,-835],[-962,-835],[3,-958],[-3,958],[3,-961],[-3,961],[-821,-818],[-824,-818],[-822,819],[-825,819],[-823,-820],[-826,-820],[-673,-670],[-676,-670],[-674,671],[-677,671],[-675,-672],[-678,-672],[682,673],[-682,-673],[-683,674],[684,675],[-684,-675],[904,800],[909,800],[908,904],[-908,-904],[1,-905],[-1,905],[1,-908],[-1,908],[-790,-967],[793,-967],[790,-968],[-793,-968],[-968,784],[-967,784],[-791,785],[-794,785],[-792,-1023],[795,-1023],[792,-1024],[-795,-1024],[-1024,786],[-1023,786],[-224,-221],[-230,-221],[-420,-221],[-234,-221],[-225,222],[-231,222],[-231,-234],[-231,-420],[-225,-234],[-225,-420],[-226,-223],[-232,-223],[-420,-223],[-234,-223],[16,-420],[-16,420],[16,-421],[-16,421],[-455,-979],[238,-979],[455,-980],[-238,-980],[-980,224],[-979,224],[-456,225],[-239,225],[-457,-1025],[240,-1025],[457,-1026],[-240,-1026],[-1026,226],[-1025,226],[274,241],[-274,-241],[-275,242],[276,243],[-276,-243],[-280,-274],[-266,275],[-278,275],[-281,275],[-282,-276],[-648,-280],[-631,-280],[-612,281],[-643,281],[-649,281],[-649,-631],[-643,-631],[-612,-631],[-650,-282],[-631,-282],[637,631],[638,631],[623,-634],[-623,634],[623,-637],[-623,637],[13,623],[-13,-623],[654,642],[660,642],[-655,643],[-661,643],[656,644],[662,644],[617,611],[620,611],[-618,612],[-621,612],[619,613],[622,613],[-501,-277],[-507,-277],[-508,-277],[-502,278],[-502,-508],[-502,-507],[-503,-279],[-507,-279],[-508,-279],[36,-504],[-36,504],[36,-507],[-36,507],[-478,266],[-272,266],[-69,-75],[-79,76],[-82,76],[-70,76],[-71,-77],[72,69],[-72,-69],[-73,70],[74,71],[-74,-71],[297,72],[-292,73],[-295,73],[-298,73],[299,74],[-691,-297],[-726,-297],[-697,-297],[-692,298],[-692,-697],[-692,-726],[-693,-299],[-726,-299],[-697,-299],[24,-726],[-24,726],[24,-727],[-24,727],[-312,-294],[-315,-294],[-313,295],[-316,295],[-314,-296],[-317,-296],[300,291],[-300,-291],[-301,292],[302,293],[-302,-293],[306,300],[309,300],[-307,301],[-310,301],[308,302],[311,302],[-715,-81],[-752,-81],[-719,-81],[-716,82],[-716,-719],[-716,-752],[-717,-83],[-752,-83],[-719,-83],[2,-752],[-2,752],[2,-753],[-2,753],[-121,79],[-236,79],[-57,-1028],[55,-1028],[57,-1029],[-55,-1029],[-1029,1027],[-1028,1027],[-198,-55],[-174,-55],[-199,56],[-175,56],[-200,-57],[-176,-57],[-189,-174],[-180,-174],[-183,-174],[-375,-174],[-190,175],[-181,175],[-184,175],[-376,175],[-191,-176],[-182,-176],[-185,-176],[-377,-176],[186,-375],[-186,375],[186,-378],[-186,378],[186,-396],[-186,396],[187,-376],[187,-379],[187,-397],[188,-377],[-188,377],[188,-380],[-188,380],[188,-398],[-188,398],[-756,-186],[-746,-186],[-750,187],[-747,187],[-747,-756],[-750,753],[-754,-756],[-750,-756],[-756,-188],[-748,-188],[759,746],[-759,-746],[-760,747],[761,748],[-761,-748],[-763,760],[-766,760],[-766,769],[-763,886],[571,183],[574,183],[-572,184],[-575,184],[573,185],[576,185],[138,-354],[-138,354],[138,-357],[-138,357],[138,-363],[-138,363],[138,-574],[-138,574],[139,-355],[139,-358],[139,-364],[139,-575],[140,-356],[-140,356],[140,-359],[-140,359],[140,-365],[-140,365],[140,-576],[-140,576],[538,138],[-538,-138],[-539,139],[540,140],[-540,-140],[-851,-538],[-855,539],[-862,539],[-852,539],[-862,865],[-855,858],[-853,-540],[868,851],[-868,-851],[-869,852],[870,853],[-870,-853],[-878,-868],[-872,869],[-872,875],[-876,-878],[-872,-878],[-878,-870],[348,-366],[-348,366],[348,-369],[-348,369],[348,-381],[-348,381],[348,-402],[-348,402],[348,-571],[-348,571],[349,-367],[349,-370],[349,-382],[349,-403],[349,-572],[350,-368],[-350,368],[350,-371],[-350,371],[350,-383],[-350,383],[350,-404],[-350,404],[350,-573],[-350,573],[733,348],[720,348],[-734,349],[-721,349],[735,350],[722,350],[-730,-720],[-724,721],[-724,727],[-728,-730],[-724,-730],[-730,-722],[-737,734],[-740,734],[-740,743],[-737,881],[393,180],[384,180],[405,180],[-394,181],[-385,181],[-406,181],[395,182],[386,182],[407,182],[387,-405],[-387,405],[387,-408],[-387,408],[388,-406],[388,-409],[389,-407],[-389,407],[389,-410],[-389,410],[414,387],[-414,-387],[-415,388],[416,389],[-416,-389],[360,-411],[-360,411],[360,-414],[-360,414],[361,-412],[361,-415],[362,-413],[-362,413],[362,-416],[-362,416],[-594,-360],[-598,361],[-605,361],[-595,361],[-605,608],[-598,601],[-596,-362],[628,594],[-625,595],[-625,-634],[-635,628],[-625,628],[628,596],[-433,-384],[-428,385],[-418,385],[-434,385],[-435,-386],[253,-430],[-253,430],[253,-433],[-253,433],[254,-431],[254,-434],[255,-432],[-255,432],[255,-435],[-255,435],[-256,-253],[-260,254],[-263,254],[-257,254],[-263,486],[-260,509],[-258,-255],[467,256],[-467,-256],[-468,257],[469,258],[-469,-258],[-490,-467],[-475,468],[-475,504],[-505,-490],[-475,-490],[-490,-469],[-424,-417],[-472,418],[-472,421],[-422,-424],[-472,-424],[-424,-419],[-437,428],[-440,428],[-440,443],[-437,772],[345,-390],[-345,390],[345,-393],[-345,393],[346,-391],[346,-394],[347,-392],[-347,392],[347,-395],[-347,395],[568,345],[541,345],[-562,346],[-542,346],[-542,568],[-562,-565],[-566,568],[-562,568],[568,347],[543,347],[-545,542],[-552,542],[-552,555],[-545,548],[177,-189],[-177,189],[177,-192],[-177,192],[178,-190],[178,-193],[179,-191],[-179,191],[179,-194],[-179,194],[-812,-177],[-828,178],[-831,178],[-813,178],[-831,934],[-828,942],[-814,-179],[815,812],[-815,-812],[-816,813],[817,814],[-817,-814],[-938,-815],[-929,816],[-929,958],[-959,-938],[-929,-938],[-938,-817],[-911,-198],[-918,-198],[-915,199],[-912,199],[-912,-918],[-915,921],[-915,-918],[-922,-918],[-913,-200],[-918,-200],[891,911],[-891,-911],[-892,912],[893,913],[-893,-913],[-895,892],[-902,892],[-902,898],[-895,905],[-51,-1031],[49,-1031],[51,-1032],[-49,-1032],[-1032,1030],[-1031,1030],[150,49],[159,49],[132,49],[144,49],[-151,50],[-160,50],[-133,50],[-145,50],[152,51],[161,51],[134,51],[146,51],[135,-144],[-135,144],[135,-147],[-135,147],[135,-351],[-135,351],[135,-399],[-135,399],[136,-145],[136,-148],[136,-352],[136,-400],[137,-146],[-137,146],[137,-149],[-137,149],[137,-353],[-137,353],[137,-401],[-137,401],[396,135],[-396,-135],[-397,136],[398,137],[-398,-137],[363,132],[411,132],[369,132],[-364,133],[-412,133],[-370,133],[365,134],[413,134],[371,134],[129,-159],[-129,159],[129,-162],[-129,162],[130,-160],[130,-163],[131,-161],[-131,161],[131,-164],[-131,164],[351,129],[357,129],[366,129],[342,129],[-352,130],[-358,130],[-367,130],[-343,130],[353,131],[359,131],[368,131],[344,131],[390,342],[-390,-342],[-391,343],[392,344],[-392,-344],[126,-150],[-126,150],[126,-153],[-126,153],[126,-171],[-126,171],[127,-151],[127,-154],[127,-172],[128,-152],[-128,152],[128,-155],[-128,155],[128,-173],[-128,173],[192,126],[-192,-126],[-193,127],[194,128],[-194,-128],[-48,-1034],[46,-1034],[48,-1035],[-46,-1035],[-1035,1033],[-1034,1033],[354,46],[165,46],[147,46],[153,46],[-355,47],[-166,47],[-148,47],[-154,47],[356,48],[167,48],[149,48],[155,48],[141,-165],[-141,165],[141,-168],[-141,168],[142,-166],[142,-169],[143,-167],[-143,167],[143,-170],[-143,170],[372,141],[-372,-141],[-373,142],[374,143],[-374,-143],[-378,-372],[-381,-372],[-379,373],[-382,373],[-380,-374],[-383,-374],[-54,-1037],[52,-1037],[54,-1038],[-52,-1038],[-1038,1036],[-1037,1036],[162,52],[168,52],[156,52],[171,52],[-163,53],[-169,53],[-157,53],[-172,53],[164,54],[170,54],[158,54],[173,54],[399,156],[430,156],[402,156],[408,156],[-400,157],[-431,157],[-403,157],[-409,157],[401,158],[432,158],[404,158],[410,158],[-1036,1039],[-1033,1039],[-1030,1039],[-1027,1039],[-1019,1039],[-994,1039],[-987,1039],[-986,41,38],[988,39,-37],[989,-39,37],[-987,989,988],[-38,37,39],[-38,-37,-39],[-202,201,203],[-202,-201,-203],[-950,949,951],[-950,-949,-951],[-915,914,916],[-915,-914,-916],[-855,854,856],[-855,-854,-856],[-839,838,840],[-839,-838,-840],[-828,827,829],[-828,-827,-829],[-779,778,780],[-779,-778,-780],[-763,762,764],[-763,-762,-764],[-737,736,738],[-737,-736,-738],[-709,708,710],[-709,-708,-710],[-640,639,641],[-640,-639,-641],[-598,597,599],[-598,-597,-599],[-545,544,546],[-545,-544,-546],[-496,495,497],[-496,-495,-497],[-437,436,438],[-437,-436,-438],[-289,288,290],[-289,-288,-290],[-260,259,261],[-260,-259,-261],[-215,214,216],[-215,-214,-216],[-208,207,209],[-208,-207,-209],[-205,204,206],[-205,-204,-206],[-63,62,64],[-63,-62,-64],[686,884,885],[65,286,68],[61,547,533],[66,857,850],[494,512,513],[926,945,946],[777,924,925],[213,775,776],[591,600,593],[702,889,890],[995,42,-40],[996,-42,40],[-994,996,995],[-41,40,42],[-41,-40,-42],[-677,676,678],[-677,-676,-678],[-661,660,662],[-661,-660,-662],[-658,657,659],[-658,-657,-659],[-196,195,197],[-196,-195,-197],[-902,901,903],[-902,-901,-903],[-862,861,863],[-862,-861,-863],[-831,830,832],[-831,-830,-832],[-825,824,826],[-825,-824,-826],[-794,793,795],[-794,-793,-795],[-766,765,767],[-766,-765,-767],[-740,739,741],[-740,-739,-741],[-621,620,622],[-621,-620,-622],[-605,604,606],[-605,-604,-606],[-552,551,553],[-552,-551,-553],[-522,521,523],[-522,-521,-523],[-440,439,441],[-440,-439,-441],[-340,339,341],[-340,-339,-341],[-325,324,326],[-325,-324,-326],[-316,315,317],[-316,-315,-317],[-310,309,311],[-310,-309,-311],[-272,271,273],[-272,-271,-273],[-263,262,264],[-263,-262,-264],[-251,250,252],[-251,-250,-252],[-239,238,240],[-239,-238,-240],[-236,235,237],[-236,-235,-237],[-465,-466,448],[-465,-464,446],[-465,446,448],[-465,-466,653],[-465,-464,651],[-465,651,653],[-465,-466,110],[-465,-464,108],[-465,108,110],[-447,-448,466],[-447,-446,464],[-447,464,466],[-447,-448,653],[-447,-446,651],[-447,651,653],[-447,-448,110],[-447,-446,108],[-447,108,110],[-652,-653,466],[-652,-651,464],[-652,464,466],[-652,-653,448],[-652,-651,446],[-652,446,448],[-652,-653,110],[-652,-651,108],[-652,108,110],[-109,-110,466],[-109,-108,464],[-109,464,466],[-109,-110,448],[-109,-108,446],[-109,446,448],[-109,-110,653],[-109,-108,651],[-109,651,653],[-88,87,89],[-88,-87,-89],[-465,464,466],[-465,-464,-466],[-269,478,481],[-481,480,482],[-481,-480,-482],[-478,477,479],[-478,-477,-479],[-269,268,270],[-269,-268,-270],[-484,499,502],[-502,501,503],[-502,-501,-503],[-499,498,500],[-499,-498,-500],[975,668,-495],[976,-668,495],[-483,976,975],[-484,483,485],[-484,-483,-485],[1001,668,-497],[1002,-668,497],[-485,1002,1001],[-680,-681,-787],[-680,-679,-789],[-680,-787,-789],[-680,-681,-458],[-680,-679,-460],[-680,-458,-460],[-788,-789,-679],[-788,-787,-681],[-788,-679,-681],[-788,-789,-458],[-788,-787,-460],[-788,-458,-460],[-459,-460,-679],[-459,-458,-681],[-459,-679,-681],[-459,-460,-787],[-459,-458,-789],[-459,-787,-789],[-447,446,448],[-447,-446,-448],[-822,821,823],[-822,-821,-823],[-683,682,684],[-683,-682,-684],[-680,679,681],[-680,-679,-681],[-453,452,454],[-453,-452,-454],[-932,953,956],[-956,955,957],[-956,-955,-957],[-953,952,954],[-953,-952,-954],[963,948,-949],[964,-948,949],[-931,964,963],[-932,931,933],[-932,-931,-933],[1003,948,-951],[1004,-948,951],[-933,1004,1003],[-450,788,791],[-791,790,792],[-791,-790,-792],[-788,787,789],[-788,-787,-789],[-450,449,451],[-450,-449,-451],[-585,807,810],[-810,809,811],[-810,-809,-811],[-807,806,808],[-807,-806,-808],[969,782,-778],[970,-782,778],[-584,970,969],[-585,584,586],[-585,-584,-586],[1005,782,-780],[1006,-782,780],[-586,1006,1005],[-218,456,459],[-459,458,460],[-459,-458,-460],[-456,455,457],[-456,-455,-457],[-218,217,219],[-218,-217,-219],[-211,228,231],[-231,230,232],[-231,-230,-232],[-228,227,229],[-228,-227,-229],[981,798,-214],[982,-798,214],[-210,982,981],[-211,210,212],[-211,-210,-212],[1007,798,-216],[1008,-798,216],[-212,1008,1007],[-462,652,655],[-655,654,656],[-655,-654,-656],[-652,651,653],[-652,-651,-653],[-462,461,463],[-462,-461,-463],[-582,615,618],[-618,617,619],[-618,-617,-619],[-615,614,616],[-615,-614,-616],[-582,581,583],[-582,-581,-583],[-589,646,649],[-649,648,650],[-649,-648,-650],[-646,645,647],[-646,-645,-647],[973,666,-639],[974,-666,639],[-588,974,973],[-589,588,590],[-589,-588,-590],[1009,666,-641],[1010,-666,641],[-590,1010,1009],[-334,-335,-303],[-334,-333,-305],[-334,-303,-305],[-334,-335,-318],[-334,-333,-320],[-334,-318,-320],[-334,-335,-123],[-334,-333,-125],[-334,-123,-125],[-304,-305,-333],[-304,-303,-335],[-304,-333,-335],[-304,-305,-318],[-304,-303,-320],[-304,-318,-320],[-304,-305,-123],[-304,-303,-125],[-304,-123,-125],[-319,-320,-333],[-319,-318,-335],[-319,-333,-335],[-319,-320,-303],[-319,-318,-305],[-319,-303,-305],[-319,-320,-123],[-319,-318,-125],[-319,-123,-125],[-124,-125,-333],[-124,-123,-335],[-124,-333,-335],[-124,-125,-303],[-124,-123,-305],[-124,-303,-305],[-124,-125,-318],[-124,-123,-320],[-124,-318,-320],[-109,108,110],[-109,-108,-110],[-519,518,520],[-519,-518,-520],[-337,336,338],[-337,-336,-338],[-334,333,335],[-334,-333,-335],[-118,117,119],[-118,-117,-119],[-59,528,531],[-531,530,532],[-531,-530,-532],[-528,527,529],[-528,-527,-529],[983,67,-204],[984,-67,204],[-58,984,983],[-59,58,60],[-59,-58,-60],[1011,67,-206],[1012,-67,206],[-60,1012,1011],[-313,312,314],[-313,-312,-314],[-307,306,308],[-307,-306,-308],[-304,303,305],[-304,-303,-305],[-115,114,116],[-115,-114,-116],[-688,692,695],[-695,694,696],[-695,-694,-696],[-692,691,693],[-692,-691,-693],[965,837,-838],[966,-837,838],[-687,966,965],[-688,687,689],[-688,-687,-689],[1013,837,-840],[1014,-837,840],[-689,1014,1013],[-112,319,322],[-322,321,323],[-322,-321,-323],[-319,318,320],[-319,-318,-320],[-112,111,113],[-112,-111,-113],[-284,843,846],[-846,845,847],[-846,-845,-847],[-843,842,844],[-843,-842,-844],[977,287,-288],[978,-287,288],[-283,978,977],[-284,283,285],[-284,-283,-285],[1015,287,-290],[1016,-287,290],[-285,1016,1015],[-85,121,124],[-124,123,125],[-124,-123,-125],[-121,120,122],[-121,-120,-122],[-85,84,86],[-85,-84,-86],[-704,713,716],[-716,715,717],[-716,-715,-717],[-713,712,714],[-713,-712,-714],[971,707,-708],[972,-707,708],[-703,972,971],[-704,703,705],[-704,-703,-705],[1017,707,-710],[1018,-707,710],[-705,1018,1017],[1020,45,-43],[1021,-45,43],[-1019,1021,1020],[-44,43,45],[-44,-43,-45],[-625,624,626],[-625,-624,-626],[-562,561,563],[-562,-561,-563],[-559,558,560],[-559,-558,-560],[-248,247,249],[-248,-247,-249],[-929,928,930],[-929,-928,-930],[-895,894,896],[-895,-894,-896],[-872,871,873],[-872,-871,-873],[-750,749,751],[-750,-749,-751],[-724,723,725],[-724,-723,-725],[-578,577,579],[-578,-577,-579],[-475,474,476],[-475,-474,-476],[-472,471,473],[-472,-471,-473],[-91,-92,107],[-91,-90,105],[-91,105,107],[-91,-92,77],[-91,-90,75],[-91,75,77],[-106,-107,92],[-106,-105,90],[-106,90,92],[-106,-107,77],[-106,-105,75],[-106,75,77],[-76,-77,92],[-76,-75,90],[-76,90,92],[-76,-77,107],[-76,-75,105],[-76,105,107],[-103,102,104],[-103,-102,-104],[-96,-93,-90],[99,93,90],[99,96,90],[-97,-98,95],[-97,-96,93],[-97,93,95],[-94,-95,98],[-94,-93,96],[-94,96,98],[-97,-98,-99],[-97,-96,-101],[-97,-99,-101],[-94,-95,-99],[-94,-93,-101],[-94,-99,-101],[-91,90,92],[-91,-90,-92],[-98,-95,-92],[101,95,92],[101,98,92],[-327,-515,-99],[-525,-526,-527],[-525,-524,-529],[-525,-527,-529],[-328,-329,517],[-328,-327,515],[-328,515,517],[-516,-517,329],[-516,-515,327],[-516,327,329],[-328,-329,-527],[-328,-327,-529],[-328,-527,-529],[-516,-517,-527],[-516,-515,-529],[-516,-527,-529],[-100,99,101],[-100,-99,-101],[-329,-517,-101],[-524,-564,-537],[-328,327,329],[-328,-327,-329],[330,336,339],[-340,-341,-336],[-340,-339,-338],[-340,-336,-338],[-337,-338,-339],[-337,-336,-341],[-337,-339,-341],[-331,330,332],[-331,-330,-332],[332,338,341],[-515,-518,-521],[-522,-523,520],[-522,-521,518],[-522,518,520],[-519,-520,523],[-519,-518,521],[-519,521,523],[-516,515,517],[-516,-515,-517],[-517,-520,-523],[-97,96,98],[-97,-96,-98],[324,321,93],[-324,-321,93],[-324,321,-93],[324,-321,-93],[-94,93,95],[-94,-93,-95],[326,323,95],[-326,-323,95],[-326,323,-95],[326,-323,-95],[-245,-246,-221],[-245,-244,-223],[-245,-221,-223],[-245,-246,-241],[-245,-244,-243],[-245,-241,-243],[-222,-223,-244],[-222,-221,-246],[-222,-244,-246],[-222,-223,-241],[-222,-221,-243],[-222,-241,-243],[-242,-243,-244],[-242,-241,-246],[-242,-244,-246],[-242,-243,-221],[-242,-241,-223],[-242,-221,-223],[-106,105,107],[-106,-105,-107],[-803,-784,-244],[-803,-809,-244],[-803,-800,-244],[-801,-802,-809],[-801,-800,-811],[-801,-809,-811],[-801,-802,-784],[-801,-800,-786],[-801,-784,-786],[-810,-811,-784],[-810,-809,-786],[-810,-784,-786],[-785,-786,-809],[-785,-784,-811],[-785,-809,-811],[-801,-802,805],[-801,-800,803],[-801,803,805],[-810,-811,805],[-810,-809,803],[-810,803,805],[-785,-786,805],[-785,-784,803],[-785,803,805],[-245,244,246],[-245,-244,-246],[-805,-786,-246],[-805,-811,-246],[-805,-800,-246],[818,670,803],[-833,-670,-803],[-833,-818,-803],[-819,-820,-670],[-819,-818,-672],[-819,-670,-672],[-671,-672,-818],[-671,-670,-820],[-671,-818,-820],[-819,-820,835],[-819,-818,833],[-819,833,835],[-671,-672,835],[-671,-670,833],[-671,833,835],[-804,803,805],[-804,-803,-805],[820,672,805],[-835,-672,-805],[-835,-820,-805],[-834,833,835],[-834,-833,-835],[818,821,824],[-825,-826,-821],[-825,-824,-823],[-825,-821,-823],[-822,-823,-824],[-822,-821,-826],[-822,-824,-826],[-819,818,820],[-819,-818,-820],[820,823,826],[670,673,676],[-677,-678,-673],[-677,-676,-675],[-677,-673,-675],[-674,-675,-676],[-674,-673,-678],[-674,-676,-678],[-671,670,672],[-671,-670,-672],[672,675,678],[-674,673,675],[-674,-673,-675],[-800,-904,-909],[967,790,-793],[968,-790,793],[-784,968,967],[-785,784,786],[-785,-784,-786],[1023,792,-795],[1024,-792,795],[-786,1024,1023],[-231,-232,-224],[-231,-230,-226],[-231,-224,-226],[-225,-226,-230],[-225,-224,-232],[-225,-230,-232],[-222,221,223],[-222,-221,-223],[979,455,-238],[980,-455,238],[-224,980,979],[-225,224,226],[-225,-224,-226],[1025,457,-240],[1026,-457,240],[-226,1026,1025],[-242,241,243],[-242,-241,-243],[-277,-265,-274],[280,265,274],[280,277,274],[-278,-279,267],[-278,-277,265],[-278,265,267],[-266,-267,279],[-266,-265,277],[-266,277,279],[-278,-279,-280],[-278,-277,-282],[-278,-280,-282],[-266,-267,-280],[-266,-265,-282],[-266,-280,-282],[-275,274,276],[-275,-274,-276],[-279,-267,-276],[282,267,276],[282,279,276],[-642,-611,-280],[-632,-633,-648],[-632,-631,-650],[-632,-648,-650],[-643,-644,613],[-643,-642,611],[-643,611,613],[-612,-613,644],[-612,-611,642],[-612,642,644],[-643,-644,-648],[-643,-642,-650],[-643,-648,-650],[-612,-613,-648],[-612,-611,-650],[-612,-648,-650],[-281,280,282],[-281,-280,-282],[-644,-613,-282],[-631,-637,-638],[-642,-654,-660],[-661,-662,656],[-661,-660,654],[-661,654,656],[-655,-656,662],[-655,-654,660],[-655,660,662],[-643,642,644],[-643,-642,-644],[-644,-656,-662],[-611,-617,-620],[-621,-622,619],[-621,-620,617],[-621,617,619],[-618,-619,622],[-618,-617,620],[-618,620,622],[-612,611,613],[-612,-611,-613],[-613,-619,-622],[-278,277,279],[-278,-277,-279],[271,477,265],[-271,-477,265],[-271,477,-265],[271,-477,-265],[-266,265,267],[-266,-265,-267],[273,479,267],[-273,-479,267],[-273,479,-267],[273,-479,-267],[-81,-78,-75],[69,78,75],[69,81,75],[-82,-83,80],[-82,-81,78],[-82,78,80],[-79,-80,83],[-79,-78,81],[-79,81,83],[-82,-83,-69],[-82,-81,-71],[-82,-69,-71],[-79,-80,-69],[-79,-78,-71],[-79,-69,-71],[-76,75,77],[-76,-75,-77],[-83,-80,-77],[71,80,77],[71,83,77],[-70,69,71],[-70,-69,-71],[294,291,72],[-297,-291,-72],[-297,-294,-72],[-295,-296,-291],[-295,-294,-293],[-295,-291,-293],[-292,-293,-294],[-292,-291,-296],[-292,-294,-296],[-295,-296,299],[-295,-294,297],[-295,297,299],[-292,-293,299],[-292,-291,297],[-292,297,299],[-73,72,74],[-73,-72,-74],[296,293,74],[-299,-293,-74],[-299,-296,-74],[-298,297,299],[-298,-297,-299],[294,312,315],[-316,-317,-312],[-316,-315,-314],[-316,-312,-314],[-313,-314,-315],[-313,-312,-317],[-313,-315,-317],[-295,294,296],[-295,-294,-296],[296,314,317],[-292,291,293],[-292,-291,-293],[-300,-306,-309],[-310,-311,308],[-310,-309,306],[-310,306,308],[-307,-308,311],[-307,-306,309],[-307,309,311],[-301,300,302],[-301,-300,-302],[-302,-308,-311],[-82,81,83],[-82,-81,-83],[235,120,78],[-235,-120,78],[-235,120,-78],[235,-120,-78],[-79,78,80],[-79,-78,-80],[237,122,80],[-237,-122,80],[-237,122,-80],[237,-122,-80],[1028,57,-55],[1029,-57,55],[-1027,1029,1028],[55,198,174],[-175,-176,-198],[-175,-174,-200],[-175,-198,-200],[-199,-200,-174],[-199,-198,-176],[-199,-174,-176],[-56,55,57],[-56,-55,-57],[57,200,176],[-376,-377,-183],[-376,-375,-185],[-376,-183,-185],[-376,-377,-180],[-376,-375,-182],[-376,-180,-182],[-376,-377,-189],[-376,-375,-191],[-376,-189,-191],[-184,-185,-375],[-184,-183,-377],[-184,-375,-377],[-184,-185,-180],[-184,-183,-182],[-184,-180,-182],[-184,-185,-189],[-184,-183,-191],[-184,-189,-191],[-181,-182,-375],[-181,-180,-377],[-181,-375,-377],[-181,-182,-183],[-181,-180,-185],[-181,-183,-185],[-181,-182,-189],[-181,-180,-191],[-181,-189,-191],[-190,-191,-375],[-190,-189,-377],[-190,-375,-377],[-190,-191,-183],[-190,-189,-185],[-190,-183,-185],[-190,-191,-180],[-190,-189,-182],[-190,-180,-182],[-175,174,176],[-175,-174,-176],[-397,396,398],[-397,-396,-398],[-379,378,380],[-379,-378,-380],[-376,375,377],[-376,-375,-377],[-753,-749,-186],[-757,-758,-746],[-757,-756,-748],[-757,-746,-748],[-754,-755,751],[-754,-753,749],[-754,749,751],[-754,-755,-746],[-754,-753,-748],[-754,-746,-748],[-750,-751,-746],[-750,-749,-748],[-750,-746,-748],[-187,186,188],[-187,-186,-188],[-753,-751,-188],[-747,746,748],[-747,-746,-748],[-886,-762,-759],[-769,-765,-759],[765,762,759],[769,762,759],[765,886,759],[769,886,759],[-770,-771,767],[-770,-769,765],[-770,765,767],[-887,-888,764],[-887,-886,762],[-887,762,764],[-760,759,761],[-760,-759,-761],[-886,-764,-761],[-769,-767,-761],[767,764,761],[769,764,761],[767,886,761],[769,886,761],[-183,-571,-574],[-575,-576,573],[-575,-574,571],[-575,571,573],[-572,-573,576],[-572,-571,574],[-572,574,576],[-184,183,185],[-184,-183,-185],[-185,-573,-576],[-575,574,576],[-575,-574,-576],[-364,363,365],[-364,-363,-365],[-358,357,359],[-358,-357,-359],[-355,354,356],[-355,-354,-356],[-139,138,140],[-139,-138,-140],[-858,-854,-538],[-865,-861,-538],[-866,-867,863],[-866,-865,861],[-866,861,863],[-859,-860,856],[-859,-858,854],[-859,854,856],[-866,-867,-851],[-866,-865,-853],[-866,-851,-853],[-862,-863,-851],[-862,-861,-853],[-862,-851,-853],[-859,-860,-851],[-859,-858,-853],[-859,-851,-853],[-855,-856,-851],[-855,-854,-853],[-855,-851,-853],[-539,538,540],[-539,-538,-540],[-858,-856,-540],[-865,-863,-540],[-852,851,853],[-852,-851,-853],[-875,-871,-868],[878,871,868],[878,875,868],[-876,-877,873],[-876,-875,871],[-876,871,873],[-869,868,870],[-869,-868,-870],[-875,-873,-870],[878,873,870],[878,875,870],[-572,571,573],[-572,-571,-573],[-403,402,404],[-403,-402,-404],[-382,381,383],[-382,-381,-383],[-370,369,371],[-370,-369,-371],[-367,366,368],[-367,-366,-368],[-348,-733,-720],[-721,-722,735],[-721,-720,733],[-721,733,735],[-734,-735,722],[-734,-733,720],[-734,720,722],[-349,348,350],[-349,-348,-350],[-350,-735,-722],[-727,-723,-720],[730,723,720],[730,727,720],[-728,-729,725],[-728,-727,723],[-728,723,725],[-721,720,722],[-721,-720,-722],[-727,-725,-722],[730,725,722],[730,727,722],[-881,-736,-733],[-743,-739,-733],[739,736,733],[743,736,733],[739,881,733],[743,881,733],[-744,-745,741],[-744,-743,739],[-744,739,741],[-882,-883,738],[-882,-881,736],[-882,736,738],[-734,733,735],[-734,-733,-735],[-881,-738,-735],[-743,-741,-735],[741,738,735],[743,738,735],[741,881,735],[743,881,735],[-406,-407,386],[-406,-405,384],[-406,384,386],[-406,-407,395],[-406,-405,393],[-406,393,395],[-385,-386,407],[-385,-384,405],[-385,405,407],[-385,-386,395],[-385,-384,393],[-385,393,395],[-394,-395,407],[-394,-393,405],[-394,405,407],[-394,-395,386],[-394,-393,384],[-394,384,386],[-181,180,182],[-181,-180,-182],[-388,406,409],[-409,408,410],[-409,-408,-410],[-406,405,407],[-406,-405,-407],[-388,387,389],[-388,-387,-389],[-361,412,415],[-415,414,416],[-415,-414,-416],[-412,411,413],[-412,-411,-413],[-601,-597,-360],[-608,-604,-360],[-609,-610,606],[-609,-608,604],[-609,604,606],[-602,-603,599],[-602,-601,597],[-602,597,599],[-609,-610,-594],[-609,-608,-596],[-609,-594,-596],[-605,-606,-594],[-605,-604,-596],[-605,-594,-596],[-602,-603,-594],[-602,-601,-596],[-602,-594,-596],[-598,-599,-594],[-598,-597,-596],[-598,-594,-596],[-361,360,362],[-361,-360,-362],[-601,-599,-362],[-608,-606,-362],[634,624,594],[-628,-624,-594],[-628,-634,-594],[-635,-636,-624],[-635,-634,-626],[-635,-624,-626],[-595,594,596],[-595,-594,-596],[634,626,596],[-628,-626,-596],[-628,-634,-596],[-417,-427,-384],[433,427,384],[433,417,384],[-418,-419,429],[-418,-417,427],[-418,427,429],[-428,-429,419],[-428,-427,417],[-428,417,419],[-418,-419,-433],[-418,-417,-435],[-418,-433,-435],[-428,-429,-433],[-428,-427,-435],[-428,-433,-435],[-385,384,386],[-385,-384,-386],[-419,-429,-386],[435,429,386],[435,419,386],[-254,431,434],[-434,433,435],[-434,-433,-435],[-431,430,432],[-431,-430,-432],[-509,-259,-253],[-486,-262,-253],[-487,-488,264],[-487,-486,262],[-487,262,264],[-510,-511,261],[-510,-509,259],[-510,259,261],[-487,-488,-256],[-487,-486,-258],[-487,-256,-258],[-263,-264,-256],[-263,-262,-258],[-263,-256,-258],[-510,-511,-256],[-510,-509,-258],[-510,-256,-258],[-260,-261,-256],[-260,-259,-258],[-260,-256,-258],[-254,253,255],[-254,-253,-255],[-509,-261,-255],[-486,-264,-255],[-257,256,258],[-257,-256,-258],[-504,-474,-467],[490,474,467],[490,504,467],[-505,-506,476],[-505,-504,474],[-505,474,476],[-468,467,469],[-468,-467,-469],[-504,-476,-469],[490,476,469],[490,504,469],[-421,-471,-417],[424,471,417],[424,421,417],[-422,-423,473],[-422,-421,471],[-422,471,473],[-418,417,419],[-418,-417,-419],[-421,-473,-419],[424,473,419],[424,421,419],[-772,-436,-427],[-443,-439,-427],[439,436,427],[443,436,427],[439,772,427],[443,772,427],[-444,-445,441],[-444,-443,439],[-444,439,441],[-773,-774,438],[-773,-772,436],[-773,436,438],[-428,427,429],[-428,-427,-429],[-772,-438,-429],[-443,-441,-429],[441,438,429],[443,438,429],[441,772,429],[443,772,429],[-346,391,394],[-394,393,395],[-394,-393,-395],[-391,390,392],[-391,-390,-392],[565,561,345],[-569,-570,543],[-569,-568,541],[-569,541,543],[-566,-567,-561],[-566,-565,-563],[-566,-561,-563],[-566,-567,543],[-566,-565,541],[-566,541,543],[-562,-563,543],[-562,-561,541],[-562,541,543],[-346,345,347],[-346,-345,-347],[565,563,347],[-548,-544,-541],[-555,-551,-541],[551,544,541],[555,544,541],[551,548,541],[555,548,541],[-556,-557,553],[-556,-555,551],[-556,551,553],[-549,-550,546],[-549,-548,544],[-549,544,546],[-542,541,543],[-542,-541,-543],[-548,-546,-543],[-555,-553,-543],[553,546,543],[555,546,543],[553,548,543],[555,548,543],[-178,190,193],[-193,192,194],[-193,-192,-194],[-190,189,191],[-190,-189,-191],[-942,-827,-177],[-934,-830,-177],[-935,-936,832],[-935,-934,830],[-935,830,832],[-943,-944,829],[-943,-942,827],[-943,827,829],[-935,-936,-812],[-935,-934,-814],[-935,-812,-814],[-831,-832,-812],[-831,-830,-814],[-831,-812,-814],[-943,-944,-812],[-943,-942,-814],[-943,-812,-814],[-828,-829,-812],[-828,-827,-814],[-828,-812,-814],[-178,177,179],[-178,-177,-179],[-942,-829,-179],[-934,-832,-179],[-813,812,814],[-813,-812,-814],[-958,-928,-815],[938,928,815],[938,958,815],[-959,-960,930],[-959,-958,928],[-959,928,930],[-816,815,817],[-816,-815,-817],[-958,-930,-817],[938,930,817],[938,958,817],[-914,-921,-198],[-919,-920,-911],[-919,-918,-913],[-919,-911,-913],[-922,-923,916],[-922,-921,914],[-922,914,916],[-915,-916,-911],[-915,-914,-913],[-915,-911,-913],[-922,-923,-911],[-922,-921,-913],[-922,-911,-913],[-199,198,200],[-199,-198,-200],[-916,-921,-200],[-912,911,913],[-912,-911,-913],[-894,-905,-891],[-901,-898,-891],[898,905,891],[901,905,891],[898,894,891],[901,894,891],[-899,-900,903],[-899,-898,901],[-899,901,903],[-906,-907,896],[-906,-905,894],[-906,894,896],[-892,891,893],[-892,-891,-893],[-896,-905,-893],[-903,-898,-893],[898,905,893],[903,905,893],[898,896,893],[903,896,893],[1031,51,-49],[1032,-51,49],[-1030,1032,1031],[-145,-146,134],[-145,-144,132],[-145,132,134],[-145,-146,161],[-145,-144,159],[-145,159,161],[-145,-146,152],[-145,-144,150],[-145,150,152],[-133,-134,146],[-133,-132,144],[-133,144,146],[-133,-134,161],[-133,-132,159],[-133,159,161],[-133,-134,152],[-133,-132,150],[-133,150,152],[-160,-161,146],[-160,-159,144],[-160,144,146],[-160,-161,134],[-160,-159,132],[-160,132,134],[-160,-161,152],[-160,-159,150],[-160,150,152],[-151,-152,146],[-151,-150,144],[-151,144,146],[-151,-152,134],[-151,-150,132],[-151,132,134],[-151,-152,161],[-151,-150,159],[-151,159,161],[-50,49,51],[-50,-49,-51],[-400,399,401],[-400,-399,-401],[-352,351,353],[-352,-351,-353],[-148,147,149],[-148,-147,-149],[-145,144,146],[-145,-144,-146],[-136,135,137],[-136,-135,-137],[-370,-371,413],[-370,-369,411],[-370,411,413],[-370,-371,365],[-370,-369,363],[-370,363,365],[-412,-413,371],[-412,-411,369],[-412,369,371],[-412,-413,365],[-412,-411,363],[-412,363,365],[-364,-365,371],[-364,-363,369],[-364,369,371],[-364,-365,413],[-364,-363,411],[-364,411,413],[-133,132,134],[-133,-132,-134],[-130,160,163],[-163,162,164],[-163,-162,-164],[-160,159,161],[-160,-159,-161],[-343,-344,368],[-343,-342,366],[-343,366,368],[-343,-344,359],[-343,-342,357],[-343,357,359],[-343,-344,353],[-343,-342,351],[-343,351,353],[-367,-368,344],[-367,-366,342],[-367,342,344],[-367,-368,359],[-367,-366,357],[-367,357,359],[-367,-368,353],[-367,-366,351],[-367,351,353],[-358,-359,344],[-358,-357,342],[-358,342,344],[-358,-359,368],[-358,-357,366],[-358,366,368],[-358,-359,353],[-358,-357,351],[-358,351,353],[-352,-353,344],[-352,-351,342],[-352,342,344],[-352,-353,368],[-352,-351,366],[-352,366,368],[-352,-353,359],[-352,-351,357],[-352,357,359],[-130,129,131],[-130,-129,-131],[-343,342,344],[-343,-342,-344],[-172,171,173],[-172,-171,-173],[-154,153,155],[-154,-153,-155],[-151,150,152],[-151,-150,-152],[-127,126,128],[-127,-126,-128],[1034,48,-46],[1035,-48,46],[-1033,1035,1034],[-154,-155,149],[-154,-153,147],[-154,147,149],[-154,-155,167],[-154,-153,165],[-154,165,167],[-154,-155,356],[-154,-153,354],[-154,354,356],[-148,-149,155],[-148,-147,153],[-148,153,155],[-148,-149,167],[-148,-147,165],[-148,165,167],[-148,-149,356],[-148,-147,354],[-148,354,356],[-166,-167,155],[-166,-165,153],[-166,153,155],[-166,-167,149],[-166,-165,147],[-166,147,149],[-166,-167,356],[-166,-165,354],[-166,354,356],[-355,-356,155],[-355,-354,153],[-355,153,155],[-355,-356,149],[-355,-354,147],[-355,147,149],[-355,-356,167],[-355,-354,165],[-355,165,167],[-47,46,48],[-47,-46,-48],[-142,166,169],[-169,168,170],[-169,-168,-170],[-166,165,167],[-166,-165,-167],[-142,141,143],[-142,-141,-143],[372,378,381],[-382,-383,-378],[-382,-381,-380],[-382,-378,-380],[-379,-380,-381],[-379,-378,-383],[-379,-381,-383],[-373,372,374],[-373,-372,-374],[374,380,383],[1037,54,-52],[1038,-54,52],[-1036,1038,1037],[-172,-173,158],[-172,-171,156],[-172,156,158],[-172,-173,170],[-172,-171,168],[-172,168,170],[-172,-173,164],[-172,-171,162],[-172,162,164],[-157,-158,173],[-157,-156,171],[-157,171,173],[-157,-158,170],[-157,-156,168],[-157,168,170],[-157,-158,164],[-157,-156,162],[-157,162,164],[-169,-170,173],[-169,-168,171],[-169,171,173],[-169,-170,158],[-169,-168,156],[-169,156,158],[-169,-170,164],[-169,-168,162],[-169,162,164],[-163,-164,173],[-163,-162,171],[-163,171,173],[-163,-164,158],[-163,-162,156],[-163,156,158],[-163,-164,170],[-163,-162,168],[-163,168,170],[-53,52,54],[-53,-52,-54],[-409,-410,404],[-409,-408,402],[-409,402,404],[-409,-410,432],[-409,-408,430],[-409,430,432],[-409,-410,401],[-409,-408,399],[-409,399,401],[-403,-404,410],[-403,-402,408],[-403,408,410],[-403,-404,432],[-403,-402,430],[-403,430,432],[-403,-404,401],[-403,-402,399],[-403,399,401],[-431,-432,410],[-431,-430,408],[-431,408,410],[-431,-432,404],[-431,-430,402],[-431,402,404],[-431,-432,401],[-431,-430,399],[-431,399,401],[-400,-401,410],[-400,-399,408],[-400,408,410],[-400,-401,404],[-400,-399,402],[-400,402,404],[-400,-401,432],[-400,-399,430],[-400,430,432],[-157,156,158],[-157,-156,-158],[669,665,667,706],[-993,855,915,950],[663,667,665,669],[-665,-796,-781,-797],[-196,658,661,677],[268,498,489,493],[270,500,489,493],[446,458,787,679],[448,460,789,681],[-453,680,683,822],[452,952,937,941],[454,954,937,941],[449,806,897,917],[451,808,897,917],[217,227,442,233],[219,229,442,233],[581,645,607,592],[583,647,607,592],[-118,334,337,519],[117,530,554,534],[119,532,554,534],[-115,304,307,313],[114,694,742,698],[116,696,742,698],[111,842,864,848],[113,844,864,848],[84,712,768,718],[86,714,768,718],[-248,559,562,625],[-102,-75,-105,-90],[-104,-77,-107,-92],[-100,99,-96,-93],[-100,101,-98,-95],[524,527,515,99],[524,527,327,99],[-525,524,-327,-515],[-525,526,-329,-517],[-528,527,-327,-515],[-528,529,-329,-517],[524,529,517,101],[524,529,329,101],[96,845,874,849],[98,847,874,849],[105,241,221,244],[107,243,223,246],[800,809,784,244],[800,811,786,246],[-834,-833,818,670],[-834,-835,820,672],[833,955,961,962],[835,957,961,962],[-281,280,-277,-265],[-281,282,-279,-267],[631,648,611,280],[631,648,642,280],[-632,631,-642,-611],[-632,633,-644,-613],[-649,648,-642,-611],[-649,650,-644,-613],[631,650,613,282],[631,650,644,282],[277,501,507,508],[279,503,507,508],[-70,69,-81,-78],[-70,71,-83,-80],[-298,-297,294,291],[-298,-299,296,293],[297,691,726,697],[299,693,726,697],[81,715,752,719],[83,717,752,719],[-187,376,379,397],[746,756,749,186],[746,756,753,186],[-747,-753,-748,-749],[-747,-753,-746,-751],[-747,-753,-749,-751],[-757,-753,-758,-749],[-757,-753,-756,-751],[-757,-753,-749,-751],[748,756,751,188],[748,756,753,188],[-770,-886,-771,-762],[-770,-886,-769,-764],[-770,-886,-762,-764],[-766,-886,-767,-762],[-766,-886,-765,-764],[-766,-886,-762,-764],[-887,-769,-888,-765],[-887,-769,-886,-767],[-887,-769,-765,-767],[-763,-769,-764,-765],[-763,-769,-762,-767],[-763,-769,-765,-767],[851,861,854,538],[851,865,854,538],[851,861,858,538],[851,865,858,538],[-852,-865,-853,-861],[-852,-865,-851,-863],[-852,-865,-861,-863],[-852,-858,-853,-854],[-852,-858,-851,-856],[-852,-858,-854,-856],[-866,-858,-867,-854],[-866,-858,-865,-856],[-866,-858,-854,-856],[-862,-858,-863,-854],[-862,-858,-861,-856],[-862,-858,-854,-856],[-859,-865,-860,-861],[-859,-865,-858,-863],[-859,-865,-861,-863],[-855,-865,-856,-861],[-855,-865,-854,-863],[-855,-865,-861,-863],[853,863,856,540],[853,865,856,540],[853,863,858,540],[853,865,858,540],[-879,-875,-880,-871],[-879,-875,-878,-873],[-879,-875,-871,-873],[-731,-727,-732,-723],[-731,-727,-730,-725],[-731,-727,-723,-725],[-744,-881,-745,-736],[-744,-881,-743,-738],[-744,-881,-736,-738],[-740,-881,-741,-736],[-740,-881,-739,-738],[-740,-881,-736,-738],[-882,-743,-883,-739],[-882,-743,-881,-741],[-882,-743,-739,-741],[-737,-743,-738,-739],[-737,-743,-736,-741],[-737,-743,-739,-741],[-180,-393,-384,-405],[-182,-395,-386,-407],[594,604,597,360],[594,608,597,360],[594,604,601,360],[594,608,601,360],[-595,-608,-596,-604],[-595,-608,-594,-606],[-595,-608,-604,-606],[-595,-601,-596,-597],[-595,-601,-594,-599],[-595,-601,-597,-599],[-609,-601,-610,-597],[-609,-601,-608,-599],[-609,-601,-597,-599],[-605,-601,-606,-597],[-605,-601,-604,-599],[-605,-601,-597,-599],[-602,-608,-603,-604],[-602,-608,-601,-606],[-602,-608,-604,-606],[-598,-608,-599,-604],[-598,-608,-597,-606],[-598,-608,-604,-606],[596,606,599,362],[596,608,599,362],[596,606,601,362],[596,608,601,362],[-629,634,-630,626],[-629,634,-628,624],[-629,634,624,626],[-434,433,-417,-427],[-434,435,-419,-429],[256,262,259,253],[256,486,259,253],[256,262,509,253],[256,486,509,253],[-257,-486,-258,-262],[-257,-486,-256,-264],[-257,-486,-262,-264],[-257,-509,-258,-259],[-257,-509,-256,-261],[-257,-509,-259,-261],[-487,-509,-488,-259],[-487,-509,-486,-261],[-487,-509,-259,-261],[-263,-509,-264,-259],[-263,-509,-262,-261],[-263,-509,-259,-261],[-510,-486,-511,-262],[-510,-486,-509,-264],[-510,-486,-262,-264],[-260,-486,-261,-262],[-260,-486,-259,-264],[-260,-486,-262,-264],[258,264,261,255],[258,486,261,255],[258,264,509,255],[258,486,509,255],[-491,-504,-492,-474],[-491,-504,-490,-476],[-491,-504,-474,-476],[-425,-421,-426,-471],[-425,-421,-424,-473],[-425,-421,-471,-473],[-444,-772,-445,-436],[-444,-772,-443,-438],[-444,-772,-436,-438],[-440,-772,-441,-436],[-440,-772,-439,-438],[-440,-772,-436,-438],[-773,-443,-774,-439],[-773,-443,-772,-441],[-773,-443,-439,-441],[-437,-443,-438,-439],[-437,-443,-436,-441],[-437,-443,-439,-441],[-541,-568,-561,-345],[-541,-568,-565,-345],[-542,565,-543,563],[-542,565,-541,561],[-542,565,561,563],[-569,565,-570,563],[-569,565,-568,561],[-569,565,561,563],[-543,-568,-563,-347],[-543,-568,-565,-347],[-556,-548,-557,-544],[-556,-548,-555,-546],[-556,-548,-544,-546],[-552,-548,-553,-544],[-552,-548,-551,-546],[-552,-548,-544,-546],[-549,-555,-550,-551],[-549,-555,-548,-553],[-549,-555,-551,-553],[-545,-555,-546,-551],[-545,-555,-544,-553],[-545,-555,-551,-553],[812,830,827,177],[812,934,827,177],[812,830,942,177],[812,934,942,177],[-813,-934,-814,-830],[-813,-934,-812,-832],[-813,-934,-830,-832],[-813,-942,-814,-827],[-813,-942,-812,-829],[-813,-942,-827,-829],[-935,-942,-936,-827],[-935,-942,-934,-829],[-935,-942,-827,-829],[-831,-942,-832,-827],[-831,-942,-830,-829],[-831,-942,-827,-829],[-943,-934,-944,-830],[-943,-934,-942,-832],[-943,-934,-830,-832],[-828,-934,-829,-830],[-828,-934,-827,-832],[-828,-934,-830,-832],[814,832,829,179],[814,934,829,179],[814,832,942,179],[814,934,942,179],[-939,-958,-940,-928],[-939,-958,-938,-930],[-939,-958,-928,-930],[918,911,921,198],[918,911,914,198],[-919,-921,-920,-914],[-919,-921,-918,-916],[-919,-921,-914,-916],[-912,-921,-913,-914],[-912,-921,-911,-916],[-912,-921,-914,-916],[918,913,921,200],[918,913,916,200],[-902,-905,-903,-894],[-902,-905,-901,-896],[-902,-905,-894,-896],[-899,-905,-900,-894],[-899,-905,-898,-896],[-899,-905,-894,-896],[-895,-898,-896,-901],[-895,-898,-894,-903],[-895,-898,-901,-903],[-906,-898,-907,-901],[-906,-898,-905,-903],[-906,-898,-901,-903],[-132,-363,-411,-369],[-134,-365,-413,-371],[-127,151,154,172],[-1040,1019,994,987],[-62,-699,-663,-65,-685],[-64,-701,-663,-65,-685],[-87,-108,-651,-446,-464],[-89,-110,-653,-448,-466],[108,123,318,303,333],[110,125,320,305,335],[-1022,750,872,895,929],[221,224,230,420,234],[223,226,232,420,234],[174,189,180,183,375],[176,191,182,185,377],[-139,355,358,364,575],[-49,-150,-159,-132,-144],[-51,-152,-161,-134,-146],[-136,145,148,352,400],[-129,-351,-357,-366,-342],[-131,-353,-359,-368,-344],[-46,-354,-165,-147,-153],[-48,-356,-167,-149,-155],[-52,-162,-168,-156,-171],[-54,-164,-170,-158,-173],[-156,-399,-430,-402,-408],[-158,-401,-432,-404,-410],[53,47,50,56,44,986],[-63,205,208,215,260,990],[-990,289,437,496,545,991],[-991,598,640,709,737,992],[-992,763,779,828,839,993],[-88,236,239,251,263,997],[-997,272,310,316,325,998],[-998,340,440,522,552,999],[-999,605,621,740,766,1000],[-1000,794,825,831,862,902],[-103,472,475,578,724,1022],[-349,367,370,382,403,572],[-1039,1036,1033,1030,1027,1040]]".toCharArray());
    }

    public static boolean CNFExample3(ConjunctiveNormalForm cnf) {
        return JSONToCNF(cnf, "[[16,17,30],[-17,22,30],[-17,-22,30],[16,-30,47],[16,-30,-47],[-16,-21,31],[-16,-21,-31],[-16,21,-28],[-13,21,28],[13,-16,18],[13,-18,-38],[13,-18,-31],[31,38,44],[-8,31,-44],[8,-12,-44],[8,12,-27],[12,27,40],[-4,27,-40],[12,23,-40],[-3,4,-23],[3,-23,-49],[3,-13,-49],[-23,-26,49],[12,-34,49],[-12,26,-34],[19,34,36],[-19,26,36],[-30,34,-36],[24,34,-36],[-24,-36,43],[6,42,-43],[-24,42,-43],[-5,-24,-42],[5,20,-42],[5,-7,-20],[4,7,10],[-4,10,-20],[7,-10,-41],[-10,41,46],[-33,41,-46],[33,-37,-46],[32,33,37],[6,-32,37],[-6,25,-32],[-6,-25,-48],[-9,28,48],[-9,-25,-28],[19,-25,48],[2,9,-19],[-2,-19,35],[-2,22,-35],[-22,-35,50],[-17,-35,-50],[-29,-35,-50],[-1,29,-50],[1,11,29],[-11,17,-45],[-11,39,45],[-26,39,45],[-3,-26,45],[-11,15,-39],[14,-15,-39],[14,-15,-45],[14,-15,-27],[-14,-15,47],[17,17,40],[1,-29,-31],[-7,32,38],[-14,-33,-47],[-1,2,-8],[35,43,44],[21,21,24],[20,29,-48],[23,35,-37],[2,18,-33],[15,25,-45],[9,14,-38],[-5,11,50],[-3,-13,46],[-13,-41,43]]".toCharArray());
    }

    public static boolean CNFExample2(ConjunctiveNormalForm cnf) {
        return JSONToCNF(cnf, "[[16,30,95],[-16,30,95],[-30,35,78],[-30,-78,85],[-78,-85,95],[8,55,100],[8,55,-95],[9,52,100],[9,73,-100],[-8,-9,52],[38,66,83],[-38,83,87],[-52,83,-87],[66,74,-83],[-52,-66,89],[-52,73,-89],[-52,73,-74],[-8,-73,-95],[40,-55,90],[-40,-55,90],[25,35,82],[-25,82,-90],[-55,-82,-90],[11,75,84],[11,-75,96],[23,-75,-96],[-11,23,-35],[-23,29,65],[29,-35,-65],[-23,-29,84],[-35,54,70],[-54,70,77],[19,-77,-84],[-19,-54,70],[22,68,81],[-22,48,81],[-22,-48,93],[3,-48,-93],[7,18,-81],[-7,56,-81],[3,18,-56],[-18,47,68],[-18,-47,-81],[-3,68,77],[-3,-77,-84],[19,-68,-70],[-19,-68,74],[-68,-70,-74],[54,61,-62],[50,53,-62],[-50,61,-62],[-27,56,93],[4,14,76],[4,-76,96],[-4,14,80],[-14,-68,80],[-10,-39,-89],[1,49,-81],[1,26,-49],[17,-26,-49],[-1,17,-40],[16,51,-89],[-9,57,60],[12,45,-51],[2,12,69],[2,-12,40],[-12,-51,69],[-33,60,-98],[5,-32,-66],[2,-47,-100],[-42,64,83],[20,-42,-64],[20,-48,98],[-20,50,98],[-32,-50,98],[-24,37,-73],[-24,-37,-100],[-57,71,81],[-37,40,-91],[31,42,81],[-31,42,72],[-31,42,-72],[7,-19,25],[-1,-25,-94],[-15,-44,79],[-6,31,46],[-39,41,88],[28,-39,43],[28,-43,-88],[-4,-28,-88],[-30,-39,-41],[-29,33,88],[-16,21,94],[-10,26,62],[-11,-64,86],[-6,-41,76],[38,-46,93],[26,-37,94],[-26,53,-79],[78,87,-94],[65,76,-87],[23,51,-62],[-11,-36,57],[41,59,-65],[-56,72,-91],[13,-20,-46],[-13,15,79],[-17,47,-60],[-13,-44,99],[-7,-38,67],[37,-49,62],[-14,-17,-79],[-13,-15,-22],[32,-33,-34],[24,45,48],[21,24,-48],[-36,64,-85],[10,-61,67],[-5,44,59],[-80,-85,-99],[6,37,-97],[-21,-34,64],[-5,44,46],[58,-76,97],[-21,-36,75],[-15,58,-59],[-58,-76,-99],[-2,15,33],[-26,34,-57],[-18,-82,-92],[27,-80,-97],[6,32,63],[-34,-86,92],[13,-61,97],[-28,43,-98],[5,39,-86],[39,-45,92],[27,-43,97],[13,-58,-86],[-28,-67,-93],[-69,85,99],[42,71,-72],[10,-27,-63],[-59,63,-83],[36,86,-96],[-2,36,75],[-59,-71,89],[36,-67,91],[36,-60,63],[-63,91,-93],[25,87,92],[-21,49,-71],[-2,10,22],[6,-18,41],[6,71,-92],[-53,-69,-71],[-2,-53,-58],[43,-45,-96],[34,-45,-69],[63,-86,-98]]".toCharArray());
    }

    public static boolean CNFExample8(ConjunctiveNormalForm cnf) {
        return JSONToCNF(cnf, "[[-2,1],[-13,1],[-20,1],[-22,1],[-26,1],[-38,1],[-55,1],[-56,1],[-57,1],[-58,1],[-59,1],[-61,1],[-1,2],[-1,13],[-1,20],[-1,22],[-1,26],[-1,38],[-1,59],[-3,2],[-4,2],[-5,2],[-6,2],[-7,2],[-8,2],[-9,2],[-11,2],[-2,-11,11],[-10,9],[-9,10],[-12,11],[-11,12],[-14,13],[-15,13],[-16,13],[-17,13],[-18,13],[-19,13],[-13,15],[-21,20],[-23,22],[-25,22],[-24,23],[-23,24],[-27,26],[-28,26],[-29,26],[-30,26],[-31,26],[-32,26],[-33,26],[-34,26],[-35,26],[-36,26],[-37,26],[-26,34],[-26,35],[-26,36],[-39,38],[-40,38],[-41,38],[-53,38],[-54,38],[-42,41],[-43,41],[-44,41],[-45,41],[-46,41],[-47,41],[-48,41],[-49,41],[-50,41],[-51,41],[-52,41],[-60,59],[-62,61],[-64,61],[-66,61],[-69,61],[-70,61],[-71,61],[-73,61],[-74,61],[-91,61],[-61,69],[-63,62],[-62,63],[-65,64],[-67,66],[-68,66],[-72,71],[-71,72],[-75,74],[-78,74],[-79,74],[-80,74],[-81,74],[-82,74],[-83,74],[-84,74],[-85,74],[-86,74],[-87,74],[-88,74],[-74,75],[-74,78],[-74,80],[-74,81],[-74,82],[-74,83],[-74,84],[-74,85],[-74,88],[-76,75],[-77,75],[-75,76],[-75,77],[-89,88],[-90,88],[-88,89],[-88,90],[-92,91],[-96,91],[-97,91],[-98,91],[-99,91],[-100,91],[-91,92],[-93,92],[-94,92],[-95,92],[-92,93],[-92,94],[-92,95],[-102,101],[-103,101],[-105,101],[-190,101],[-223,101],[-233,101],[-104,103],[-103,104],[-106,105],[-107,105],[-108,105],[-109,105],[-110,105],[-119,105],[-131,105],[-143,105],[-174,105],[-175,105],[-179,105],[-180,105],[-105,175],[-111,110],[-114,110],[-115,110],[-116,110],[-117,110],[-118,110],[-110,116],[-110,117],[-110,118],[-112,111],[-113,111],[-111,112],[-111,113],[-120,119],[-121,119],[-122,119],[-123,119],[-125,119],[-127,119],[-129,119],[-119,121],[-124,123],[-123,124],[-126,125],[-125,126],[-128,127],[-127,128],[-130,129],[-129,130],[-132,131],[-133,131],[-134,131],[-136,131],[-138,131],[-140,131],[-142,131],[-135,134],[-134,135],[-137,136],[-136,137],[-139,138],[-138,139],[-141,140],[-140,141],[-144,143],[-153,143],[-145,144],[-146,144],[-147,144],[-150,144],[-144,150],[-148,147],[-149,147],[-147,148],[-147,149],[-151,150],[-152,150],[-150,151],[-150,152],[-154,153],[-155,153],[-162,153],[-169,153],[-172,153],[-153,169],[-156,155],[-157,155],[-158,155],[-159,155],[-160,155],[-161,155],[-155,156],[-155,157],[-155,158],[-155,159],[-155,160],[-155,161],[-163,162],[-164,162],[-165,162],[-166,162],[-167,162],[-168,162],[-162,163],[-162,164],[-162,165],[-162,166],[-162,167],[-162,168],[-170,169],[-171,169],[-169,170],[-169,171],[-173,172],[-172,173],[-176,175],[-177,175],[-178,175],[-175,176],[-175,177],[-175,178],[-181,180],[-182,180],[-183,180],[-184,180],[-185,180],[-186,180],[-187,180],[-188,180],[-189,180],[-191,190],[-192,190],[-193,190],[-195,190],[-196,190],[-203,190],[-207,190],[-208,190],[-209,190],[-211,190],[-215,190],[-190,195],[-190,209],[-194,193],[-193,194],[-197,196],[-198,196],[-199,198],[-200,198],[-201,198],[-202,198],[-204,203],[-205,203],[-206,203],[-203,205],[-210,209],[-209,210],[-212,211],[-213,211],[-211,212],[-211,213],[-214,213],[-213,214],[-216,215],[-217,215],[-218,217],[-219,217],[-220,217],[-221,217],[-217,218],[-217,219],[-217,220],[-222,221],[-221,222],[-224,223],[-225,223],[-226,223],[-229,223],[-223,229],[-227,226],[-228,226],[-226,227],[-226,228],[-230,229],[-231,229],[-232,229],[-229,230],[-229,231],[-229,232],[-234,233],[-233,234],[-235,234],[-236,234],[-237,234],[-238,234],[-239,234],[-234,235],[-234,236],[-234,238],[-241,240],[-261,240],[-263,240],[-264,240],[-265,240],[-266,240],[-267,240],[-268,240],[-269,240],[-270,240],[-271,240],[-272,240],[-240,261],[-240,-241,266],[-240,272],[-242,241],[-247,241],[-250,241],[-251,241],[-252,241],[-253,241],[-254,241],[-260,241],[-243,242],[-244,242],[-245,242],[-246,242],[-248,247],[-249,247],[-255,254],[-256,254],[-257,254],[-258,254],[-259,254],[-254,-247,255],[-262,261],[-273,272],[-274,272],[-275,272],[-276,272],[-277,272],[-272,273],[-272,274],[-272,275],[-272,276],[-272,277],[-278,277],[-277,278],[-280,279],[-288,279],[-291,279],[-307,279],[-308,279],[-320,279],[-344,279],[-359,279],[-379,279],[-381,279],[-383,279],[-279,280],[-279,291],[-279,308],[-279,320],[-279,344],[-279,379],[-279,381],[-279,383],[-281,280],[-282,281],[-283,281],[-285,281],[-287,281],[-284,283],[-286,285],[-285,286],[-289,288],[-290,288],[-292,291],[-293,291],[-294,291],[-296,291],[-297,291],[-298,291],[-301,291],[-304,291],[-291,297],[-291,298],[-295,294],[-299,298],[-300,298],[-298,299],[-302,301],[-303,301],[-301,302],[-305,304],[-306,304],[-309,308],[-310,308],[-311,308],[-312,308],[-313,308],[-315,308],[-316,308],[-317,308],[-318,308],[-319,308],[-308,-309,310],[-308,318],[-308,319],[-314,313],[-313,314],[-321,320],[-322,320],[-323,320],[-324,320],[-325,320],[-328,320],[-330,320],[-337,320],[-340,320],[-341,320],[-342,320],[-343,320],[-320,340],[-320,341],[-320,342],[-326,325],[-327,325],[-325,327],[-329,328],[-331,330],[-332,330],[-330,331],[-330,332],[-333,332],[-334,332],[-335,332],[-336,332],[-332,333],[-332,334],[-332,335],[-332,336],[-338,337],[-339,337],[-337,338],[-345,344],[-353,344],[-354,344],[-355,344],[-356,344],[-357,344],[-358,344],[-344,355],[-346,345],[-347,345],[-349,345],[-350,345],[-351,345],[-352,345],[-345,350],[-348,347],[-347,348],[-360,359],[-361,359],[-362,359],[-363,359],[-364,359],[-365,359],[-366,359],[-367,359],[-368,359],[-369,359],[-370,359],[-371,359],[-372,359],[-373,359],[-374,359],[-375,359],[-376,359],[-359,361],[-377,376],[-378,376],[-380,379],[-382,381],[-384,383],[-385,383],[-386,383],[-383,384],[-383,385],[-383,386],[-388,387],[-402,387],[-403,387],[-404,387],[-405,387],[-412,387],[-413,387],[-387,388],[-387,412],[-387,413],[-389,388],[-391,388],[-394,388],[-400,388],[-388,389],[-388,391],[-388,394],[-388,400],[-390,389],[-392,391],[-393,391],[-395,394],[-396,394],[-397,394],[-398,394],[-399,394],[-394,399],[-401,400],[-406,405],[-407,405],[-408,405],[-409,405],[-410,405],[-411,405],[-405,409],[-414,413],[-415,413],[-416,413],[-413,414],[-413,415],[-413,416],[-418,417],[-421,417],[-424,417],[-429,417],[-432,417],[-453,417],[-481,417],[-496,417],[-515,417],[-521,417],[-526,417],[-527,417],[-533,417],[-534,417],[-535,417],[-536,417],[-544,417],[-547,417],[-549,417],[-552,417],[-555,417],[-560,417],[-565,417],[-572,417],[-575,417],[-578,417],[-583,417],[-588,417],[-595,417],[-417,418],[-417,421],[-417,424],[-417,429],[-417,432],[-417,453],[-417,481],[-417,496],[-417,515],[-417,521],[-417,527],[-417,536],[-417,544],[-417,547],[-417,549],[-417,552],[-417,555],[-417,560],[-417,565],[-417,572],[-417,575],[-417,578],[-417,583],[-417,588],[-417,595],[-419,418],[-420,418],[-422,421],[-423,421],[-425,424],[-426,424],[-427,424],[-428,424],[-430,429],[-431,429],[-433,432],[-434,432],[-435,432],[-436,432],[-437,432],[-438,432],[-439,432],[-440,432],[-441,432],[-442,432],[-443,432],[-444,432],[-445,432],[-446,432],[-447,432],[-448,432],[-449,432],[-450,432],[-451,432],[-452,432],[-454,453],[-455,453],[-456,453],[-457,453],[-458,453],[-459,453],[-460,453],[-461,453],[-462,453],[-463,453],[-464,453],[-465,453],[-466,453],[-467,453],[-468,453],[-469,453],[-470,453],[-471,453],[-472,453],[-473,453],[-474,453],[-475,453],[-476,453],[-477,453],[-478,453],[-479,453],[-480,453],[-482,481],[-483,481],[-484,481],[-485,481],[-486,481],[-487,481],[-488,481],[-489,481],[-490,481],[-491,481],[-492,481],[-493,481],[-494,481],[-495,481],[-497,496],[-498,496],[-499,496],[-500,496],[-501,496],[-502,496],[-503,496],[-504,496],[-505,496],[-506,496],[-507,496],[-508,496],[-509,496],[-510,496],[-511,496],[-512,496],[-513,496],[-514,496],[-516,515],[-517,515],[-518,515],[-519,515],[-520,515],[-522,521],[-523,521],[-524,521],[-525,521],[-528,527],[-529,527],[-530,527],[-531,527],[-532,527],[-537,536],[-538,536],[-539,536],[-540,536],[-541,536],[-542,536],[-543,536],[-545,544],[-546,544],[-548,547],[-550,549],[-551,549],[-553,552],[-554,552],[-556,555],[-557,555],[-558,555],[-559,555],[-561,560],[-562,560],[-563,560],[-564,560],[-566,565],[-567,565],[-568,565],[-569,565],[-570,565],[-571,565],[-573,572],[-574,572],[-576,575],[-577,575],[-579,578],[-580,578],[-581,578],[-582,578],[-584,583],[-585,583],[-586,583],[-587,583],[-589,588],[-590,588],[-591,588],[-592,588],[-593,588],[-594,588],[-596,595],[-597,595],[-595,596],[-595,597],[-599,598],[-613,598],[-618,598],[-640,598],[-665,598],[-685,598],[-697,598],[-600,599],[-605,599],[-606,599],[-607,599],[-608,599],[-609,599],[-599,600],[-599,607],[-599,609],[-601,600],[-602,600],[-603,600],[-604,600],[-610,609],[-611,609],[-612,609],[-609,610],[-609,611],[-609,612],[-614,613],[-613,614],[-615,614],[-616,614],[-617,614],[-614,615],[-614,616],[-614,617],[-619,618],[-620,618],[-621,618],[-628,618],[-629,618],[-630,618],[-632,618],[-633,618],[-634,618],[-635,618],[-636,618],[-618,620],[-618,621],[-618,628,629],[-618,634],[-618,636],[-622,621],[-623,621],[-624,621],[-627,621],[-625,624],[-626,624],[-624,626],[-631,630],[-630,631],[-637,636],[-638,636],[-639,636],[-636,637],[-636,638],[-636,639],[-641,640],[-642,640],[-643,640],[-644,640],[-645,640],[-646,640],[-650,640],[-651,640],[-652,640],[-654,640],[-655,640],[-658,640],[-659,640],[-660,640],[-661,640],[-640,642],[-640,643],[-640,644],[-640,645],[-640,650],[-640,654],[-640,655],[-640,661],[-647,646],[-648,646],[-649,646],[-646,647],[-653,652],[-656,655],[-657,655],[-662,661],[-663,661],[-664,661],[-661,662],[-661,663],[-661,664],[-666,665],[-670,665],[-677,665],[-678,665],[-679,665],[-680,665],[-681,665],[-665,666],[-665,670],[-665,679],[-665,680],[-665,681],[-667,666],[-668,666],[-669,666],[-671,670],[-672,670],[-673,670],[-674,670],[-675,670],[-676,670],[-670,672],[-670,673],[-682,681],[-683,681],[-684,681],[-681,682],[-681,683],[-681,684],[-686,685],[-687,685],[-688,685],[-689,685],[-692,685],[-693,685],[-685,689],[-685,693],[-690,689],[-691,689],[-689,691],[-694,693],[-695,693],[-696,693],[-693,694],[-693,695],[-693,696],[-698,697],[-699,697],[-700,697],[-701,697],[-702,697],[-703,697],[-707,697],[-718,697],[-719,697],[-697,703],[-697,707],[-697,718],[-697,719],[-704,703],[-705,703],[-706,703],[-703,704],[-703,705],[-703,706],[-708,707],[-709,707],[-710,707],[-711,707],[-712,707],[-713,707],[-714,707],[-715,707],[-716,707],[-717,707],[-720,719],[-721,719],[-722,719],[-719,720],[-719,721],[-719,722],[-724,723],[-732,723],[-735,723],[-738,723],[-739,723],[-723,724],[-723,732],[-723,738],[-723,739],[-725,724],[-726,724],[-727,724],[-729,724],[-730,724],[-731,724],[-728,727],[-727,728],[-733,732],[-734,732],[-736,735],[-737,735],[-735,-733,736],[-735,-734,737],[-740,739],[-741,739],[-742,739],[-739,740],[-739,741],[-739,742],[-744,743],[-747,743],[-749,743],[-763,743],[-764,743],[-765,743],[-766,743],[-770,743],[-771,743],[-774,743],[-743,774],[-745,744],[-746,744],[-748,747],[-750,749],[-751,749],[-755,749],[-761,749],[-752,751],[-753,751],[-754,751],[-751,752],[-751,753],[-751,754],[-756,755],[-757,755],[-758,755],[-759,755],[-760,755],[-762,761],[-761,762],[-767,766],[-768,766],[-769,766],[-766,767],[-772,771],[-773,771],[-771,772],[-771,773],[-776,775],[-777,775],[-778,775],[-779,775],[-780,775],[-781,775],[-782,775],[-783,775],[-789,775],[-793,775],[-797,775],[-805,775],[-811,775],[-817,775],[-818,775],[-820,775],[-822,775],[-825,775],[-837,775],[-775,797],[-775,822],[-775,825],[-775,837],[-784,783],[-785,783],[-787,783],[-783,784],[-786,785],[-785,786],[-788,787],[-787,788],[-790,789],[-791,789],[-789,790],[-792,791],[-791,792],[-794,793],[-795,793],[-793,794],[-796,795],[-795,796],[-798,797],[-799,797],[-800,797],[-802,797],[-803,797],[-804,797],[-797,798],[-797,799],[-797,802],[-797,803],[-797,804],[-801,800],[-800,801],[-806,805],[-807,805],[-809,805],[-810,805],[-805,806],[-805,809],[-805,810],[-808,807],[-807,808],[-812,811],[-813,811],[-815,811],[-816,811],[-811,812],[-811,815],[-811,816],[-814,813],[-813,814],[-819,818],[-818,819],[-821,820],[-820,821],[-823,822],[-824,822],[-822,824],[-826,825],[-827,825],[-828,825],[-829,825],[-830,825],[-835,825],[-836,825],[-825,826],[-825,827],[-825,828],[-825,829],[-825,830],[-825,835],[-825,836],[-831,830],[-832,830],[-833,830],[-834,830],[-830,831],[-830,832],[-830,833],[-830,834],[-838,837],[-839,837],[-840,837],[-837,838],[-837,839],[-837,840],[-842,841],[-843,841],[-844,841],[-847,841],[-848,841],[-849,841],[-850,841],[-841,844],[-841,850],[-845,844],[-846,844],[-851,850],[-852,850],[-853,850],[-850,851],[-850,852],[-850,853],[-855,854],[-856,854],[-857,854],[-858,854],[-859,854],[-860,854],[-864,854],[-868,854],[-872,854],[-854,-857,858],[-854,-858,858],[-854,860],[-854,864],[-861,860],[-862,860],[-863,860],[-865,864],[-866,864],[-867,864],[-864,865],[-864,866],[-864,867],[-869,868],[-868,869],[-870,869],[-871,869],[-869,870],[-869,871],[-873,872],[-874,872],[-872,873],[-872,874],[-876,875],[-879,875],[-880,875],[-875,880],[-877,876],[-878,876],[-876,878],[-881,880],[-882,880],[-880,881],[-880,882],[-884,883],[-885,883],[-886,883],[-887,883],[-888,883],[-889,883],[-890,883],[-891,883],[-892,883],[-893,883],[-894,883],[-895,883],[-897,883],[-898,883],[-899,883],[-900,883],[-901,883],[-883,888],[-883,889],[-883,890],[-883,891],[-883,892],[-883,893],[-883,897],[-883,901],[-896,895],[-895,896],[-903,902],[-904,902],[-905,902],[-906,902],[-907,902],[-908,902],[-913,902],[-921,902],[-932,902],[-936,902],[-938,902],[-939,902],[-940,902],[-943,902],[-947,902],[-948,902],[-961,902],[-962,902],[-975,902],[-1006,902],[-1020,902],[-1027,902],[-1041,902],[-1047,902],[-902,940],[-902,943],[-902,-947,-217,948],[-909,908],[-910,908],[-911,908],[-912,908],[-914,913],[-915,913],[-917,913],[-918,913],[-920,913],[-913,914],[-913,917],[-916,915],[-915,916],[-919,918],[-918,919],[-922,921],[-923,921],[-924,921],[-925,921],[-926,921],[-927,921],[-928,921],[-930,921],[-931,921],[-921,-922,923],[-921,-922,924],[-921,925],[-921,931],[-929,928],[-928,929],[-933,932],[-934,933],[-935,933],[-933,934],[-933,935],[-937,936],[-941,940],[-942,940],[-940,941],[-940,942],[-944,943],[-945,943],[-946,943],[-943,945],[-949,948],[-950,948],[-951,948],[-954,948],[-955,948],[-952,951],[-953,951],[-956,955],[-957,955],[-958,955],[-959,955],[-960,955],[-955,956],[-955,957],[-955,958],[-955,959],[-955,960],[-963,962],[-964,962],[-965,962],[-968,962],[-969,962],[-966,965],[-967,965],[-970,969],[-971,969],[-972,969],[-973,969],[-974,969],[-969,970],[-969,971],[-969,972],[-969,973],[-969,974],[-976,975],[-977,975],[-983,975],[-984,975],[-985,975],[-986,975],[-987,975],[-988,975],[-989,975],[-990,975],[-991,975],[-992,975],[-993,975],[-996,975],[-999,975],[-1001,975],[-1005,975],[-975,983],[-975,984],[-975,987],[-975,988],[-975,989],[-975,990],[-975,991],[-975,992],[-975,993],[-975,996],[-975,1001],[-975,1005],[-978,977],[-979,977],[-980,977],[-981,977],[-982,977],[-994,993],[-995,993],[-993,994],[-993,995],[-997,996],[-998,996],[-996,997],[-996,998],[-1000,999],[-999,1000],[-1002,1001],[-1003,1001],[-1004,1001],[-1001,1003],[-1001,1004],[-1007,1006],[-1012,1006],[-1006,1007],[-1006,1012],[-1008,1007],[-1009,1007],[-1010,1007],[-1011,1007],[-1007,1008],[-1007,1009],[-1007,1010],[-1007,1011],[-1013,1012],[-1014,1012],[-1015,1012],[-1016,1012],[-1012,1013],[-1012,1014],[-1012,1016],[-1017,1016],[-1018,1016],[-1019,1016],[-1016,1019],[-1021,1020],[-1022,1020],[-1024,1020],[-1020,1024],[-1023,1022],[-1025,1024],[-1026,1024],[-1024,1025],[-1024,1026],[-1028,1027],[-1029,1027],[-1030,1027],[-1031,1027],[-1032,1027],[-1033,1027],[-1034,1027],[-1035,1027],[-1036,1027],[-1037,1027],[-1038,1027],[-1027,1028],[-1027,1029],[-1027,1030],[-1027,1031],[-1027,1032],[-1027,1033],[-1027,1035],[-1027,1037],[-1027,1038],[-1039,1038],[-1040,1038],[-1038,1039],[-1038,1040],[-1042,1041],[-1043,1041],[-1046,1041],[-1041,1042],[-1041,1043],[-1041,1046],[-1044,1043],[-1045,1043],[-1043,1044],[-1043,1045],[-1048,1047],[-1049,1047],[-1050,1047],[-1053,1047],[-1054,1047],[-1055,1047],[-1056,1047],[-1060,1047],[-1064,1047],[-1066,1047],[-1069,1047],[-1072,1047],[-1073,1047],[-1074,1047],[-1075,1047],[-1077,1047],[-1047,1048],[-1047,1050],[-1047,1053],[-1047,1064],[-1047,1066],[-1047,1069],[-1047,1072],[-1047,1074],[-1051,1050],[-1052,1050],[-1050,1051],[-1050,1052],[-1057,1056],[-1058,1056],[-1059,1056],[-1061,1060],[-1062,1060],[-1063,1060],[-1060,1061],[-1060,1062],[-1060,1063],[-1065,1064],[-1067,1066],[-1068,1066],[-1070,1069],[-1071,1069],[-1069,1070],[-1069,1071],[-1076,1075],[-1075,1076],[-1079,1078],[-1080,1078],[-1082,1078],[-1084,1078],[-1085,1078],[-1086,1078],[-1089,1078],[-1090,1078],[-1091,1078],[-1078,1084],[-1078,-909,1085],[-1078,1086],[-1078,-1079,1089],[-1081,1080],[-1080,1081],[-1083,1082],[-1082,1083],[-1087,1086],[-1088,1086],[-1086,1087],[-1086,1088],[-1093,1092],[-1092,1093],[-1095,1094],[-1094,1095],[-1097,1096],[-1099,1096],[-1102,1096],[-1096,1099],[-1096,1102],[-1098,1097],[-1100,1099],[-1101,1099],[-1099,1100],[-1099,1101],[-1104,1103],[-1105,1103],[-1106,1103],[-1107,1103],[-1108,1103],[-1109,1103],[-1110,1103],[-1111,1103],[-1112,1103],[-1113,1103],[-1118,1103],[-1119,1103],[-1142,1103],[-1103,1142],[-1114,1113],[-1115,1113],[-1116,1113],[-1117,1113],[-1113,1115],[-1120,1119],[-1131,1119],[-1121,1120],[-1122,1120],[-1123,1120],[-1126,1120],[-1129,1120],[-1120,1121],[-1124,1123],[-1125,1123],[-1123,1124],[-1123,-1124,1125],[-1127,1126],[-1128,1126],[-1126,1127],[-1126,-1128,1128],[-1130,1129],[-1129,1130],[-1132,1131],[-1133,1131],[-1134,1131],[-1137,1131],[-1140,1131],[-1131,1132],[-1135,1134],[-1136,1134],[-1134,1135],[-1134,-1135,1136],[-1138,1137],[-1139,1137],[-1137,1138],[-1137,-1139,1139],[-1141,1140],[-1140,1141],[-1143,1142],[-1144,1142],[-1145,1142],[-1142,1143],[-1142,1144],[-1142,1145],[-1146,1145],[-1147,1145],[-1145,1146],[-1145,1147],[-1149,1148],[-1150,1148],[-1152,1148],[-1161,1148],[-1162,1148],[-1163,1148],[-1148,1161],[-1148,1163],[-1151,1150],[-1150,1151],[-1153,1152],[-1154,1152],[-1155,1152],[-1156,1152],[-1157,1152],[-1152,1153],[-1152,1154],[-1152,1155],[-1152,1156],[-1158,1157],[-1159,1157],[-1160,1157],[-1157,-1158,1159],[-1157,1160],[-1165,1164],[-1166,1164],[-1167,1164],[-1168,1167],[-1167,1168],[-1170,1169],[-1174,1169],[-1175,1169],[-1176,1169],[-1177,1169],[-1182,1169],[-1183,1169],[-1184,1169],[-1185,1169],[-1186,1169],[-1169,1176],[-1169,1182],[-1169,1183],[-1169,1184],[-1169,1186],[-1171,1170],[-1172,1170],[-1173,1170],[-1170,1171],[-1170,1172],[-1170,1173],[-1178,1177],[-1179,1177],[-1180,1177],[-1181,1177],[-1177,1181],[-1188,1187],[-1187,1188],[-1190,1189],[-1191,1189],[-1192,1189],[-1193,1189],[-1194,1189],[-1199,1189],[-1189,1194],[-1189,1199],[-1195,1194],[-1196,1194],[-1197,1194],[-1198,1194],[-1194,1195],[-1194,1196],[-1194,1197],[-1194,1198],[-1201,1200],[-1202,1200],[-1203,1200],[-1204,1200],[-1205,1200],[-1208,1200],[-1219,1200],[-1222,1200],[-1225,1200],[-1200,1201],[-1200,1202],[-1200,1203],[-1200,1204],[-1200,1205],[-1200,1208],[-1200,1225],[-1206,1205],[-1207,1206],[-1209,1208],[-1210,1208],[-1211,1208],[-1212,1208],[-1213,1208],[-1214,1208],[-1215,1208],[-1216,1208],[-1217,1208],[-1218,1208],[-1208,1209],[-1208,1210],[-1208,1211],[-1220,1219],[-1221,1219],[-1219,1220],[-1219,1221],[-1223,1222],[-1224,1222],[-1222,1223],[-1222,1224],[-1227,1226],[-1228,1226],[-1229,1226],[-1226,1229],[-1230,1229],[-1231,1229],[-1229,1230],[-1229,1231],[-1233,1232],[-1235,1234],[-1236,1234],[-1237,1234],[-1238,1234],[-1239,1234],[-1240,1234],[-1241,1234],[-1244,1234],[-1234,1235],[-1234,1236],[-1234,1237],[-1234,1244],[-1242,1241],[-1243,1241],[-1241,1242],[-1241,1243],[-1,240],[-3,279],[-4,288],[-5,635],[-7,6],[-8,7],[-9,52],[-9,-42],[-11,11],[-28,-52,51],[-28,-21],[-28,-17],[-28,-54],[-28,27],[-30,28],[-30,29],[-32,-30],[-32,-31],[-32,54,28],[-33,53,380],[-37,28],[-41,39],[-42,54],[-45,51],[-45,50],[-45,49],[-45,47],[-46,-45],[-53,-65],[-53,78],[-54,78],[-57,28],[-57],[-62,-65],[-71],[-76,75],[-77,75],[-87,78],[-97,-86],[-97],[-98,62],[-99,309],[-99],[-101,875],[-103,-883],[-105,875],[-105,101],[-109,108],[-110,111,114],[-114,109],[-114,115],[-114,115],[-131,460],[-131,427],[-131,425],[-131,883],[-131,417],[-133,132],[-142,518],[-142,516],[-144,875],[-144,154],[-144,105],[-153,875],[-153,91],[-153,105],[-172,155],[-179,885],[-179,883],[-196,494],[-196,902],[-203,492],[-203,-902],[-206],[-207,-902],[-207],[-211,216],[-211,190],[-215,190],[-226,279],[-242,250],[-242,251,252,253,254],[-242,251,-251,253,254],[-242,251,-253,253,254],[-242,251,-254,253,254],[-242,-252,252,253,254],[-242,-252,-251,253,254],[-242,-252,-253,253,254],[-242,-252,-254,253,254],[-242,-253,252,253,254],[-242,-253,-251,253,254],[-242,-253,-253,253,254],[-242,-253,-254,253,254],[-242,-254,252,253,254],[-242,-254,-251,253,254],[-242,-254,-253,253,254],[-242,-254,-254,253,254],[-242,251,252,-251,254],[-242,251,-251,-251,254],[-242,251,-253,-251,254],[-242,251,-254,-251,254],[-242,-252,252,-251,254],[-242,-252,-251,-251,254],[-242,-252,-253,-251,254],[-242,-252,-254,-251,254],[-242,-253,252,-251,254],[-242,-253,-251,-251,254],[-242,-253,-253,-251,254],[-242,-253,-254,-251,254],[-242,-254,252,-251,254],[-242,-254,-251,-251,254],[-242,-254,-253,-251,254],[-242,-254,-254,-251,254],[-242,251,252,-252,254],[-242,251,-251,-252,254],[-242,251,-253,-252,254],[-242,251,-254,-252,254],[-242,-252,252,-252,254],[-242,-252,-251,-252,254],[-242,-252,-253,-252,254],[-242,-252,-254,-252,254],[-242,-253,252,-252,254],[-242,-253,-251,-252,254],[-242,-253,-253,-252,254],[-242,-253,-254,-252,254],[-242,-254,252,-252,254],[-242,-254,-251,-252,254],[-242,-254,-253,-252,254],[-242,-254,-254,-252,254],[-242,251,252,-254,254],[-242,251,-251,-254,254],[-242,251,-253,-254,254],[-242,251,-254,-254,254],[-242,-252,252,-254,254],[-242,-252,-251,-254,254],[-242,-252,-253,-254,254],[-242,-252,-254,-254,254],[-242,-253,252,-254,254],[-242,-253,-251,-254,254],[-242,-253,-253,-254,254],[-242,-253,-254,-254,254],[-242,-254,252,-254,254],[-242,-254,-251,-254,254],[-242,-254,-253,-254,254],[-242,-254,-254,-254,254],[-242,251,252,253,-251],[-242,251,-251,253,-251],[-242,251,-253,253,-251],[-242,251,-254,253,-251],[-242,-252,252,253,-251],[-242,-252,-251,253,-251],[-242,-252,-253,253,-251],[-242,-252,-254,253,-251],[-242,-253,252,253,-251],[-242,-253,-251,253,-251],[-242,-253,-253,253,-251],[-242,-253,-254,253,-251],[-242,-254,252,253,-251],[-242,-254,-251,253,-251],[-242,-254,-253,253,-251],[-242,-254,-254,253,-251],[-242,251,252,-251,-251],[-242,251,-251,-251,-251],[-242,251,-253,-251,-251],[-242,251,-254,-251,-251],[-242,-252,252,-251,-251],[-242,-252,-251,-251,-251],[-242,-252,-253,-251,-251],[-242,-252,-254,-251,-251],[-242,-253,252,-251,-251],[-242,-253,-251,-251,-251],[-242,-253,-253,-251,-251],[-242,-253,-254,-251,-251],[-242,-254,252,-251,-251],[-242,-254,-251,-251,-251],[-242,-254,-253,-251,-251],[-242,-254,-254,-251,-251],[-242,251,252,-252,-251],[-242,251,-251,-252,-251],[-242,251,-253,-252,-251],[-242,251,-254,-252,-251],[-242,-252,252,-252,-251],[-242,-252,-251,-252,-251],[-242,-252,-253,-252,-251],[-242,-252,-254,-252,-251],[-242,-253,252,-252,-251],[-242,-253,-251,-252,-251],[-242,-253,-253,-252,-251],[-242,-253,-254,-252,-251],[-242,-254,252,-252,-251],[-242,-254,-251,-252,-251],[-242,-254,-253,-252,-251],[-242,-254,-254,-252,-251],[-242,251,252,-254,-251],[-242,251,-251,-254,-251],[-242,251,-253,-254,-251],[-242,251,-254,-254,-251],[-242,-252,252,-254,-251],[-242,-252,-251,-254,-251],[-242,-252,-253,-254,-251],[-242,-252,-254,-254,-251],[-242,-253,252,-254,-251],[-242,-253,-251,-254,-251],[-242,-253,-253,-254,-251],[-242,-253,-254,-254,-251],[-242,-254,252,-254,-251],[-242,-254,-251,-254,-251],[-242,-254,-253,-254,-251],[-242,-254,-254,-254,-251],[-242,251,252,253,-252],[-242,251,-251,253,-252],[-242,251,-253,253,-252],[-242,251,-254,253,-252],[-242,-252,252,253,-252],[-242,-252,-251,253,-252],[-242,-252,-253,253,-252],[-242,-252,-254,253,-252],[-242,-253,252,253,-252],[-242,-253,-251,253,-252],[-242,-253,-253,253,-252],[-242,-253,-254,253,-252],[-242,-254,252,253,-252],[-242,-254,-251,253,-252],[-242,-254,-253,253,-252],[-242,-254,-254,253,-252],[-242,251,252,-251,-252],[-242,251,-251,-251,-252],[-242,251,-253,-251,-252],[-242,251,-254,-251,-252],[-242,-252,252,-251,-252],[-242,-252,-251,-251,-252],[-242,-252,-253,-251,-252],[-242,-252,-254,-251,-252],[-242,-253,252,-251,-252],[-242,-253,-251,-251,-252],[-242,-253,-253,-251,-252],[-242,-253,-254,-251,-252],[-242,-254,252,-251,-252],[-242,-254,-251,-251,-252],[-242,-254,-253,-251,-252],[-242,-254,-254,-251,-252],[-242,251,252,-252,-252],[-242,251,-251,-252,-252],[-242,251,-253,-252,-252],[-242,251,-254,-252,-252],[-242,-252,252,-252,-252],[-242,-252,-251,-252,-252],[-242,-252,-253,-252,-252],[-242,-252,-254,-252,-252],[-242,-253,252,-252,-252],[-242,-253,-251,-252,-252],[-242,-253,-253,-252,-252],[-242,-253,-254,-252,-252],[-242,-254,252,-252,-252],[-242,-254,-251,-252,-252],[-242,-254,-253,-252,-252],[-242,-254,-254,-252,-252],[-242,251,252,-254,-252],[-242,251,-251,-254,-252],[-242,251,-253,-254,-252],[-242,251,-254,-254,-252],[-242,-252,252,-254,-252],[-242,-252,-251,-254,-252],[-242,-252,-253,-254,-252],[-242,-252,-254,-254,-252],[-242,-253,252,-254,-252],[-242,-253,-251,-254,-252],[-242,-253,-253,-254,-252],[-242,-253,-254,-254,-252],[-242,-254,252,-254,-252],[-242,-254,-251,-254,-252],[-242,-254,-253,-254,-252],[-242,-254,-254,-254,-252],[-242,251,252,253,-253],[-242,251,-251,253,-253],[-242,251,-253,253,-253],[-242,251,-254,253,-253],[-242,-252,252,253,-253],[-242,-252,-251,253,-253],[-242,-252,-253,253,-253],[-242,-252,-254,253,-253],[-242,-253,252,253,-253],[-242,-253,-251,253,-253],[-242,-253,-253,253,-253],[-242,-253,-254,253,-253],[-242,-254,252,253,-253],[-242,-254,-251,253,-253],[-242,-254,-253,253,-253],[-242,-254,-254,253,-253],[-242,251,252,-251,-253],[-242,251,-251,-251,-253],[-242,251,-253,-251,-253],[-242,251,-254,-251,-253],[-242,-252,252,-251,-253],[-242,-252,-251,-251,-253],[-242,-252,-253,-251,-253],[-242,-252,-254,-251,-253],[-242,-253,252,-251,-253],[-242,-253,-251,-251,-253],[-242,-253,-253,-251,-253],[-242,-253,-254,-251,-253],[-242,-254,252,-251,-253],[-242,-254,-251,-251,-253],[-242,-254,-253,-251,-253],[-242,-254,-254,-251,-253],[-242,251,252,-252,-253],[-242,251,-251,-252,-253],[-242,251,-253,-252,-253],[-242,251,-254,-252,-253],[-242,-252,252,-252,-253],[-242,-252,-251,-252,-253],[-242,-252,-253,-252,-253],[-242,-252,-254,-252,-253],[-242,-253,252,-252,-253],[-242,-253,-251,-252,-253],[-242,-253,-253,-252,-253],[-242,-253,-254,-252,-253],[-242,-254,252,-252,-253],[-242,-254,-251,-252,-253],[-242,-254,-253,-252,-253],[-242,-254,-254,-252,-253],[-242,251,252,-254,-253],[-242,251,-251,-254,-253],[-242,251,-253,-254,-253],[-242,251,-254,-254,-253],[-242,-252,252,-254,-253],[-242,-252,-251,-254,-253],[-242,-252,-253,-254,-253],[-242,-252,-254,-254,-253],[-242,-253,252,-254,-253],[-242,-253,-251,-254,-253],[-242,-253,-253,-254,-253],[-242,-253,-254,-254,-253],[-242,-254,252,-254,-253],[-242,-254,-251,-254,-253],[-242,-254,-253,-254,-253],[-242,-254,-254,-254,-253],[-247,250],[-247,251,252,253,254],[-247,251,-251,253,254],[-247,251,-253,253,254],[-247,251,-254,253,254],[-247,-252,252,253,254],[-247,-252,-251,253,254],[-247,-252,-253,253,254],[-247,-252,-254,253,254],[-247,-253,252,253,254],[-247,-253,-251,253,254],[-247,-253,-253,253,254],[-247,-253,-254,253,254],[-247,-254,252,253,254],[-247,-254,-251,253,254],[-247,-254,-253,253,254],[-247,-254,-254,253,254],[-247,251,252,-251,254],[-247,251,-251,-251,254],[-247,251,-253,-251,254],[-247,251,-254,-251,254],[-247,-252,252,-251,254],[-247,-252,-251,-251,254],[-247,-252,-253,-251,254],[-247,-252,-254,-251,254],[-247,-253,252,-251,254],[-247,-253,-251,-251,254],[-247,-253,-253,-251,254],[-247,-253,-254,-251,254],[-247,-254,252,-251,254],[-247,-254,-251,-251,254],[-247,-254,-253,-251,254],[-247,-254,-254,-251,254],[-247,251,252,-252,254],[-247,251,-251,-252,254],[-247,251,-253,-252,254],[-247,251,-254,-252,254],[-247,-252,252,-252,254],[-247,-252,-251,-252,254],[-247,-252,-253,-252,254],[-247,-252,-254,-252,254],[-247,-253,252,-252,254],[-247,-253,-251,-252,254],[-247,-253,-253,-252,254],[-247,-253,-254,-252,254],[-247,-254,252,-252,254],[-247,-254,-251,-252,254],[-247,-254,-253,-252,254],[-247,-254,-254,-252,254],[-247,251,252,-254,254],[-247,251,-251,-254,254],[-247,251,-253,-254,254],[-247,251,-254,-254,254],[-247,-252,252,-254,254],[-247,-252,-251,-254,254],[-247,-252,-253,-254,254],[-247,-252,-254,-254,254],[-247,-253,252,-254,254],[-247,-253,-251,-254,254],[-247,-253,-253,-254,254],[-247,-253,-254,-254,254],[-247,-254,252,-254,254],[-247,-254,-251,-254,254],[-247,-254,-253,-254,254],[-247,-254,-254,-254,254],[-247,251,252,253,-251],[-247,251,-251,253,-251],[-247,251,-253,253,-251],[-247,251,-254,253,-251],[-247,-252,252,253,-251],[-247,-252,-251,253,-251],[-247,-252,-253,253,-251],[-247,-252,-254,253,-251],[-247,-253,252,253,-251],[-247,-253,-251,253,-251],[-247,-253,-253,253,-251],[-247,-253,-254,253,-251],[-247,-254,252,253,-251],[-247,-254,-251,253,-251],[-247,-254,-253,253,-251],[-247,-254,-254,253,-251],[-247,251,252,-251,-251],[-247,251,-251,-251,-251],[-247,251,-253,-251,-251],[-247,251,-254,-251,-251],[-247,-252,252,-251,-251],[-247,-252,-251,-251,-251],[-247,-252,-253,-251,-251],[-247,-252,-254,-251,-251],[-247,-253,252,-251,-251],[-247,-253,-251,-251,-251],[-247,-253,-253,-251,-251],[-247,-253,-254,-251,-251],[-247,-254,252,-251,-251],[-247,-254,-251,-251,-251],[-247,-254,-253,-251,-251],[-247,-254,-254,-251,-251],[-247,251,252,-252,-251],[-247,251,-251,-252,-251],[-247,251,-253,-252,-251],[-247,251,-254,-252,-251],[-247,-252,252,-252,-251],[-247,-252,-251,-252,-251],[-247,-252,-253,-252,-251],[-247,-252,-254,-252,-251],[-247,-253,252,-252,-251],[-247,-253,-251,-252,-251],[-247,-253,-253,-252,-251],[-247,-253,-254,-252,-251],[-247,-254,252,-252,-251],[-247,-254,-251,-252,-251],[-247,-254,-253,-252,-251],[-247,-254,-254,-252,-251],[-247,251,252,-254,-251],[-247,251,-251,-254,-251],[-247,251,-253,-254,-251],[-247,251,-254,-254,-251],[-247,-252,252,-254,-251],[-247,-252,-251,-254,-251],[-247,-252,-253,-254,-251],[-247,-252,-254,-254,-251],[-247,-253,252,-254,-251],[-247,-253,-251,-254,-251],[-247,-253,-253,-254,-251],[-247,-253,-254,-254,-251],[-247,-254,252,-254,-251],[-247,-254,-251,-254,-251],[-247,-254,-253,-254,-251],[-247,-254,-254,-254,-251],[-247,251,252,253,-252],[-247,251,-251,253,-252],[-247,251,-253,253,-252],[-247,251,-254,253,-252],[-247,-252,252,253,-252],[-247,-252,-251,253,-252],[-247,-252,-253,253,-252],[-247,-252,-254,253,-252],[-247,-253,252,253,-252],[-247,-253,-251,253,-252],[-247,-253,-253,253,-252],[-247,-253,-254,253,-252],[-247,-254,252,253,-252],[-247,-254,-251,253,-252],[-247,-254,-253,253,-252],[-247,-254,-254,253,-252],[-247,251,252,-251,-252],[-247,251,-251,-251,-252],[-247,251,-253,-251,-252],[-247,251,-254,-251,-252],[-247,-252,252,-251,-252],[-247,-252,-251,-251,-252],[-247,-252,-253,-251,-252],[-247,-252,-254,-251,-252],[-247,-253,252,-251,-252],[-247,-253,-251,-251,-252],[-247,-253,-253,-251,-252],[-247,-253,-254,-251,-252],[-247,-254,252,-251,-252],[-247,-254,-251,-251,-252],[-247,-254,-253,-251,-252],[-247,-254,-254,-251,-252],[-247,251,252,-252,-252],[-247,251,-251,-252,-252],[-247,251,-253,-252,-252],[-247,251,-254,-252,-252],[-247,-252,252,-252,-252],[-247,-252,-251,-252,-252],[-247,-252,-253,-252,-252],[-247,-252,-254,-252,-252],[-247,-253,252,-252,-252],[-247,-253,-251,-252,-252],[-247,-253,-253,-252,-252],[-247,-253,-254,-252,-252],[-247,-254,252,-252,-252],[-247,-254,-251,-252,-252],[-247,-254,-253,-252,-252],[-247,-254,-254,-252,-252],[-247,251,252,-254,-252],[-247,251,-251,-254,-252],[-247,251,-253,-254,-252],[-247,251,-254,-254,-252],[-247,-252,252,-254,-252],[-247,-252,-251,-254,-252],[-247,-252,-253,-254,-252],[-247,-252,-254,-254,-252],[-247,-253,252,-254,-252],[-247,-253,-251,-254,-252],[-247,-253,-253,-254,-252],[-247,-253,-254,-254,-252],[-247,-254,252,-254,-252],[-247,-254,-251,-254,-252],[-247,-254,-253,-254,-252],[-247,-254,-254,-254,-252],[-247,251,252,253,-253],[-247,251,-251,253,-253],[-247,251,-253,253,-253],[-247,251,-254,253,-253],[-247,-252,252,253,-253],[-247,-252,-251,253,-253],[-247,-252,-253,253,-253],[-247,-252,-254,253,-253],[-247,-253,252,253,-253],[-247,-253,-251,253,-253],[-247,-253,-253,253,-253],[-247,-253,-254,253,-253],[-247,-254,252,253,-253],[-247,-254,-251,253,-253],[-247,-254,-253,253,-253],[-247,-254,-254,253,-253],[-247,251,252,-251,-253],[-247,251,-251,-251,-253],[-247,251,-253,-251,-253],[-247,251,-254,-251,-253],[-247,-252,252,-251,-253],[-247,-252,-251,-251,-253],[-247,-252,-253,-251,-253],[-247,-252,-254,-251,-253],[-247,-253,252,-251,-253],[-247,-253,-251,-251,-253],[-247,-253,-253,-251,-253],[-247,-253,-254,-251,-253],[-247,-254,252,-251,-253],[-247,-254,-251,-251,-253],[-247,-254,-253,-251,-253],[-247,-254,-254,-251,-253],[-247,251,252,-252,-253],[-247,251,-251,-252,-253],[-247,251,-253,-252,-253],[-247,251,-254,-252,-253],[-247,-252,252,-252,-253],[-247,-252,-251,-252,-253],[-247,-252,-253,-252,-253],[-247,-252,-254,-252,-253],[-247,-253,252,-252,-253],[-247,-253,-251,-252,-253],[-247,-253,-253,-252,-253],[-247,-253,-254,-252,-253],[-247,-254,252,-252,-253],[-247,-254,-251,-252,-253],[-247,-254,-253,-252,-253],[-247,-254,-254,-252,-253],[-247,251,252,-254,-253],[-247,251,-251,-254,-253],[-247,251,-253,-254,-253],[-247,251,-254,-254,-253],[-247,-252,252,-254,-253],[-247,-252,-251,-254,-253],[-247,-252,-253,-254,-253],[-247,-252,-254,-254,-253],[-247,-253,252,-254,-253],[-247,-253,-251,-254,-253],[-247,-253,-253,-254,-253],[-247,-253,-254,-254,-253],[-247,-254,252,-254,-253],[-247,-254,-251,-254,-253],[-247,-254,-253,-254,-253],[-247,-254,-254,-254,-253],[-255,247],[-256,247],[-257,247],[-258,247],[-259,247],[-262,775],[-262,775],[-266,241],[-267,-464],[-269,-494],[-270],[-271],[-282,282],[-282,283,285],[-282,283,-283],[-282,-285,285],[-282,-285,-283],[-287,18],[-288,4],[-292,292],[-292,296,294],[-292,296,-296],[-292,-294,294],[-292,-294,-296],[-295,-253],[-295,-252],[-295,247],[-296,-307],[-301,309],[-301,-296],[-307,62],[-309,281],[-310,309],[-311,311],[-311,312,313],[-311,312,-312],[-311,-313,313],[-311,-313,-312],[-316,309],[-317,316],[-321,309],[-325,242],[-325,241],[-326,323],[-330,324],[-343,298],[-345,-293],[-345,294],[-348,347],[-356,321],[-364,291],[-365,320],[-366,280],[-367,344],[-368,344],[-369,344],[-370,344],[-371,344],[-372,309],[-373,309],[-374,309],[-380,33],[-380,323],[-390,279],[-392,279],[-395,242],[-396,279],[-396,279],[-398,492],[-398,417],[-401,279],[-402,279],[-405,417],[-408,408],[-408,411,410],[-408,411,-411],[-408,-410,410],[-408,-410,-411],[-410,393],[-410,409],[-411,409],[-458,-675,-674],[-458,-675,-676],[-458,-674,-675],[-458,-674,-676],[-458,-676,-675],[-458,-676,-674],[-558,556],[-559,556],[-599,417],[-601,478],[-605,423],[-606,330],[-606,279],[-608,423],[-613,523],[-613,417],[-618,622],[-618,623,624,627],[-618,623,-623,627],[-618,623,-627,627],[-618,-624,624,627],[-618,-624,-623,627],[-618,-624,-627,627],[-618,-627,624,627],[-618,-627,-623,627],[-618,-627,-627,627],[-618,623,624,-623],[-618,623,-623,-623],[-618,623,-627,-623],[-618,-624,624,-623],[-618,-624,-623,-623],[-618,-624,-627,-623],[-618,-627,624,-623],[-618,-627,-623,-623],[-618,-627,-627,-623],[-618,623,624,-624],[-618,623,-623,-624],[-618,623,-627,-624],[-618,-624,624,-624],[-618,-624,-623,-624],[-618,-624,-627,-624],[-618,-627,624,-624],[-618,-627,-623,-624],[-618,-627,-627,-624],[-618,417],[-624,279],[-627,619],[-629,-628],[-632,439],[-633,279],[-635,624,627],[-635,624,749],[-635,5],[-640,454],[-640,492],[-640,494],[-640,482],[-640,427],[-640,425],[-640,417],[-640,436],[-640,450],[-640,446],[-640,444],[-640,442],[-640,438],[-640,448],[-640,440],[-640,434],[-640,599],[-640,101],[-649,648],[-649,460],[-652,460],[-653,660],[-656,723],[-657,456],[-657,723],[-658,678],[-659,279],[-660,648],[-660,646],[-660,883],[-665,494],[-665,422],[-665,417],[-667,475],[-668,477],[-669,455],[-671,330],[-677,723],[-685,417],[-685,487],[-685,495],[-685,493],[-685,489],[-688,422],[-688,491],[-690,330],[-690,279],[-692,460],[-697,474],[-697,476],[-697,417],[-697,512],[-697,506],[-698,309],[-698,321],[-699,854],[-700,854],[-701,490],[-714,701],[-715,701],[-716,701],[-717,701],[-725,725],[-725,726,727],[-725,726,-726],[-725,-727,727],[-725,-727,-726],[-727,598],[-731,640],[-731,-726],[-733,330],[-733,-726],[-734,330],[-735,247],[-736,733],[-737,734],[-743,564],[-743,540],[-743,567],[-743,300],[-743,425],[-743,427],[-743,875],[-743,417],[-743,279],[-745,346],[-746,347],[-747,763],[-748,294],[-750,562],[-750,538],[-750,324],[-750,-306],[-750,305],[-750,304],[-750,303],[-750,747],[-756,330],[-756,301],[-757,756],[-757,294],[-761,627],[-761,618],[-763,309],[-763,279],[-763,499],[-763,510],[-763,504],[-764,509],[-764,770],[-764,749],[-764,309],[-764,279],[-764,508],[-764,502],[-765,554],[-766,763],[-766,563],[-766,494],[-766,427],[-766,460],[-766,279],[-769,770],[-770,519],[-770,517],[-770,525],[-770,764],[-770,749],[-770,288],[-771,548],[-775,279],[-776,776],[-776,778,777],[-776,778,-778],[-776,-777,777],[-776,-777,-778],[-777,-353],[-777,321],[-777,-301],[-777,294],[-777,309],[-805,387],[-811,387],[-817,309],[-818,309],[-820,309],[-841,843],[-841,845,846],[-841,845,-845],[-841,-846,846],[-841,-846,-845],[-845,309],[-856,856],[-856,863,861,862],[-856,863,-863,862],[-856,863,-862,862],[-856,-861,861,862],[-856,-861,-863,862],[-856,-861,-862,862],[-856,-862,861,862],[-856,-862,-863,862],[-856,-862,-862,862],[-856,863,861,-863],[-856,863,-863,-863],[-856,863,-862,-863],[-856,-861,861,-863],[-856,-861,-863,-863],[-856,-861,-862,-863],[-856,-862,861,-863],[-856,-862,-863,-863],[-856,-862,-862,-863],[-856,863,861,-861],[-856,863,-863,-861],[-856,863,-862,-861],[-856,-861,861,-861],[-856,-861,-863,-861],[-856,-861,-862,-861],[-856,-862,861,-861],[-856,-862,-863,-861],[-856,-862,-862,-861],[-858,857,858],[-861,279],[-868,863],[-868,854],[-872,868],[-872,74],[-872,854],[-875,426],[-876,428],[-877,330],[-879,483],[-883,570],[-883,568],[-883,535],[-883,494],[-883,425],[-883,427],[-883,875],[-883,417],[-885,279],[-886,101],[-887,279],[-887,902],[-895,460],[-899,500],[-902,904],[-902,975],[-902,594],[-902,592],[-902,490],[-902,460],[-902,425],[-902,427],[-902,482],[-902,505],[-902,417],[-902,101],[-908,905],[-909,192],[-909,906],[-910,907],[-913,908],[-915,441],[-921,908],[-921,191],[-923,922],[-924,922],[-926,922],[-927,922],[-928,922],[-929,928],[-930,922],[-932,909],[-933,743],[-944,949,963],[-948,947],[-948,217],[-949,949],[-949,955,950,951],[-949,955,-955,951],[-949,955,-951,951],[-949,-950,950,951],[-949,-950,-955,951],[-949,-950,-951,951],[-949,-951,950,951],[-949,-951,-955,951],[-949,-951,-951,951],[-949,955,950,-955],[-949,955,-955,-955],[-949,955,-951,-955],[-949,-950,950,-955],[-949,-950,-955,-955],[-949,-950,-951,-955],[-949,-951,950,-955],[-949,-951,-955,-955],[-949,-951,-951,-955],[-949,955,950,-950],[-949,955,-955,-950],[-949,955,-951,-950],[-949,-950,950,-950],[-949,-950,-955,-950],[-949,-950,-951,-950],[-949,-951,950,-950],[-949,-951,-955,-950],[-949,-951,-951,-950],[-952,921],[-954,909],[-962],[-963,963],[-963,965,969,964],[-963,965,-965,964],[-963,965,-964,964],[-963,-969,969,964],[-963,-969,-965,964],[-963,-969,-964,964],[-963,-964,969,964],[-963,-964,-965,964],[-963,-964,-964,964],[-963,965,969,-965],[-963,965,-965,-965],[-963,965,-964,-965],[-963,-969,969,-965],[-963,-969,-965,-965],[-963,-969,-964,-965],[-963,-964,969,-965],[-963,-964,-965,-965],[-963,-964,-964,-965],[-963,965,969,-969],[-963,965,-965,-969],[-963,965,-964,-969],[-963,-969,969,-969],[-963,-969,-965,-969],[-963,-969,-964,-969],[-963,-964,969,-969],[-963,-964,-965,-969],[-963,-964,-964,-969],[-966,921],[-968,909],[-975,546],[-975,883],[-975,490],[-975,460],[-975,425],[-975,427],[-975,482],[-975,505],[-975,417],[-975,101],[-977,908],[-978,909],[-979,1189],[-980,979],[-980,909],[-1006,1020],[-1006,902],[-1006,425],[-1006,427],[-1006,460],[-1006,494],[-1006,441],[-1006,454],[-1006,494],[-1006,101],[-1018,1015],[-1020,764],[-1020,902],[-1020,425],[-1020,427],[-1020,460],[-1020,494],[-1020,441],[-1020,454],[-1020,492],[-1020,494],[-1020,101],[-1022,884],[-1022,883],[-1027,902],[-1027,425],[-1027,427],[-1027,494],[-1027,441],[-1027,454],[-1027,101],[-1041,422],[-1041,425],[-1041,427],[-1041,494],[-1041,441],[-1041,454],[-1041,492],[-1041,494],[-1041,387],[-1041,902],[-1041,101],[-1047,763],[-1047,110],[-1047,902],[-1047,106],[-1047,143],[-1047,105],[-1047,101],[-1079,339],[-1079,494],[-1079,460],[-1079,422],[-1079,902],[-1079,590],[-1080,1079],[-1082,1079],[-1085,909],[-1089,1079],[-1094,-307],[-1094,279],[-1096,700],[-1096,699],[-1096,697],[-1096,902],[-1098,921],[-1103,494],[-1103,417],[-1104,1104],[-1109,-1111],[-1109,1108],[-1113,1114],[-1113,101],[-1116,41],[-1119,101],[-1122,1122],[-1122,1126,1129,1123],[-1122,1126,-1126,1123],[-1122,1126,-1123,1123],[-1122,-1129,1129,1123],[-1122,-1129,-1126,1123],[-1122,-1129,-1123,1123],[-1122,-1123,1129,1123],[-1122,-1123,-1126,1123],[-1122,-1123,-1123,1123],[-1122,1126,1129,-1126],[-1122,1126,-1126,-1126],[-1122,1126,-1123,-1126],[-1122,-1129,1129,-1126],[-1122,-1129,-1126,-1126],[-1122,-1129,-1123,-1126],[-1122,-1123,1129,-1126],[-1122,-1123,-1126,-1126],[-1122,-1123,-1123,-1126],[-1122,1126,1129,-1129],[-1122,1126,-1126,-1129],[-1122,1126,-1123,-1129],[-1122,-1129,1129,-1129],[-1122,-1129,-1126,-1129],[-1122,-1129,-1123,-1129],[-1122,-1123,1129,-1129],[-1122,-1123,-1126,-1129],[-1122,-1123,-1123,-1129],[-1125,1124],[-1128,1128],[-1133,1133],[-1133,1137,1140,1134],[-1133,1137,-1137,1134],[-1133,1137,-1134,1134],[-1133,-1140,1140,1134],[-1133,-1140,-1137,1134],[-1133,-1140,-1134,1134],[-1133,-1134,1140,1134],[-1133,-1134,-1137,1134],[-1133,-1134,-1134,1134],[-1133,1137,1140,-1137],[-1133,1137,-1137,-1137],[-1133,1137,-1134,-1137],[-1133,-1140,1140,-1137],[-1133,-1140,-1137,-1137],[-1133,-1140,-1134,-1137],[-1133,-1134,1140,-1137],[-1133,-1134,-1137,-1137],[-1133,-1134,-1134,-1137],[-1133,1137,1140,-1140],[-1133,1137,-1137,-1140],[-1133,1137,-1134,-1140],[-1133,-1140,1140,-1140],[-1133,-1140,-1137,-1140],[-1133,-1140,-1134,-1140],[-1133,-1134,1140,-1140],[-1133,-1134,-1137,-1140],[-1133,-1134,-1134,-1140],[-1136,1135],[-1139,1139],[-1148,1149],[-1148,1152,1150],[-1148,1152,-1152],[-1148,-1150,1150],[-1148,-1150,-1152],[-1148,425],[-1148,427],[-1148,875],[-1148,417],[-1148,883],[-1150,460],[-1150,-1152],[-1152,-1150],[-1159,1158],[-1162,900],[-1164,425],[-1164,427],[-1164,417],[-1164,883],[-1166,900],[-1167,1165],[-1167,482],[-1167,441],[-1167,622],[-1169,1092],[-1169,1187],[-1169,895],[-1169,1113],[-1169,425],[-1169,427],[-1169,875],[-1169,417],[-1169,460],[-1169,1103],[-1169,883],[-1170,279],[-1175],[-1178,1189],[-1180,1177],[-1185,900],[-1189,1092],[-1189,417],[-1192,387],[-1192,-1190],[-1193,439],[-1193,441],[-1193,494],[-1193,447],[-1200,902],[-1200,101],[-1219,883],[-1222,883],[-1226,875],[-1226,101],[-1232,875],[-1232,1187],[-1232,387],[-1232,417],[-1234,1232],[-1234,387],[-1234,425],[-1234,427],[-1234,417],[-1234,883],[-1238,241],[-1238,242],[-1240,900],[-2,6],[-26,27],[-26,29],[-26,31],[-38,39],[-38,40],[-1,55],[-1,56],[-61,70],[-91,100],[-105,107],[-105,108],[-110,115],[-131,132],[-153,154],[-180,181],[-180,182],[-180,183],[-180,184],[-180,185],[-180,186],[-180,187],[-180,188],[-180,189],[-190,191],[-190,192],[208,-190,-196,208],[208,-190,-207,208],[-196,-190,-196,208],[-196,-190,-207,208],[-207,-190,-196,208],[-207,-190,-207,208],[-215,216],[-241,250],[-281,282],[-291,292],[-291,293],[-308,311],[-345,352],[-405,408],[-418,419],[-418,420],[-421,422],[-424,425],[-424,427],[-429,430],[433,-432,-640,433],[-640,-432,-640,433],[435,-432,-640,435],[-640,-432,-640,435],[437,-432,-640,437],[-640,-432,-640,437],[439,-432,-640,439],[-640,-432,-640,439],[441,-432,-640,441],[-640,-432,-640,441],[443,-432,-640,443],[-640,-432,-640,443],[445,-432,-640,445],[-640,-432,-640,445],[447,-432,-640,447],[-640,-432,-640,447],[449,-432,-640,449],[-640,-432,-640,449],[451,-432,-653,451],[-653,-432,-653,451],[-453,454],[-453,456],[-453,458],[-453,460],[-453,462],[-453,464],[-453,466],[-453,468],[-453,470],[-453,472],[-453,474],[-453,476],[-453,479],[482,-481,-879,482],[-879,-481,-879,482],[486,-481,-685,486],[-685,-481,-685,486],[488,-481,-685,488],[-685,-481,-685,488],[490,-481,-688,490],[-688,-481,-688,490],[492,-481,-685,492],[-685,-481,-685,492],[494,-481,-685,494],[-685,-481,-685,494],[501,-496,-764,501],[-764,-496,-764,501],[503,-496,-763,503],[-763,-496,-763,503],[505,-496,-697,505],[-697,-496,-697,505],[507,-496,-764,507],[-764,-496,-764,507],[509,-496,-763,509],[-763,-496,-763,509],[511,-496,-697,511],[-697,-496,-697,511],[-515,516],[-515,518],[-515,520],[-521,522],[-521,524],[-527,529],[-527,531],[-417,534],[-536,537],[-536,539],[-544,545],[-549,550],[-552,553],[-555,556],[561,-560,-750,561],[-750,-560,-750,561],[563,-560,-744,563],[-744,-560,-744,563],[-565,566],[-572,573],[-575,576],[579,-578,-764,579],[-764,-578,-764,579],[581,-578,-763,581],[-763,-578,-763,581],[584,-583,-885,584],[-885,-583,-885,584],[589,-588,-1079,589],[-1079,-588,-1079,589],[591,-588,-902,591],[-902,-588,-902,591],[593,-588,-902,593],[-902,-588,-902,593],[601,-600,-478,-602,601],[601,-600,-478,-603,601],[601,-600,-478,-604,601],[-602,-600,-478,-602,601],[-602,-600,-478,-603,601],[-602,-600,-478,-604,601],[-603,-600,-478,-602,601],[-603,-600,-478,-603,601],[-603,-600,-478,-604,601],[-604,-600,-478,-602,601],[-604,-600,-478,-603,601],[-604,-600,-478,-604,601],[619,-618,-761,619],[-761,-618,-761,619],[-621,622],[628,-618,-623,628],[-623,-618,-623,628],[-724,725],[-775,776],[-841,842],[-841,843],[-841,847],[-854,855],[-854,856],[-854,857],[-883,884],[-902,903],[-902,904],[-902,905],[-902,906],[-902,907],[947,-902,-217,947],[-217,-902,-217,947],[-948,949],[-962,963],[-975,976],[-1103,1104],[1107,-1103,-1108,1107],[-1108,-1103,-1108,1107],[1114,-1113,-1117,1114],[1114,-1113,-1116,1114],[-1117,-1113,-1117,1114],[-1117,-1113,-1116,1114],[-1116,-1113,-1117,1114],[-1116,-1113,-1116,1114],[-1120,1122],[-1131,1133],[-1148,1149],[-1189,1190],[-1208,1212]]".toCharArray());
    }
}
